"""Test module for the pod reading."""

import datetime as dt
import sys
import unittest
from unittest import mock

import numpy as np
import numpy.testing

from pygac.clock_offsets_converter import txt as clock_offsets_txt
from pygac.gac_pod import GACPODReader
from pygac.lac_pod import LACPODReader
from pygac.reader import NoTLEData, ReaderError
from pygac.tests.utils import CalledWithArray


class TestPOD(unittest.TestCase):
    """Test the POD GAC reader."""

    longMessage = True

    def setUp(self):
        """Set up the test."""
        self.reader = GACPODReader()
        # python 2 compatibility
        if sys.version_info.major < 3:
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test__validate_header(self):
        """Test the header validation"""
        filename = b"NSS.GHRR.TN.D80001.S0332.E0526.B0627173.WI"
        head = {"data_set_name": filename}
        GACPODReader._validate_header(head)
        # wrong name pattern
        with self.assertRaisesRegex(ReaderError,
                                    "Data set name .* does not match!"):
            head = {"data_set_name": b"abc.txt"}
            GACPODReader._validate_header(head)
        # wrong platform
        name = b"NSS.GHRR.NL.D02187.S1904.E2058.B0921517.GC"
        with self.assertRaisesRegex(ReaderError,
                                    'Improper platform id "NL"!'):
            head = {"data_set_name": name}
            GACPODReader._validate_header(head)
        # wrong transfer mode
        name = filename.replace(b"GHRR", b"LHRR")
        with self.assertRaisesRegex(ReaderError,
                                    'Improper transfer mode "LHRR"!'):
            head = {"data_set_name": name}
            GACPODReader._validate_header(head)
        # other change reader
        head = {"data_set_name": name}
        LACPODReader._validate_header(head)

    @mock.patch("pygac.reader.Reader.get_calibrated_channels")
    def test__get_calibrated_channels_uniform_shape(self, get_channels):
        """Test the uniform shape as required by gac_io.save_gac."""
        channels = np.arange(2*2*5, dtype=float).reshape((2, 2, 5))
        get_channels.return_value = channels
        uniform_channels = self.reader._get_calibrated_channels_uniform_shape()
        self.assertTrue(np.isnan(uniform_channels[:, :, 2]).all())
        self.assertTrue(uniform_channels[:, :, [0, 1, 3, 4, 5]].sum()
                        == channels.sum())

    def test_decode_timestamps(self):
        """Test POD timestamp decoding."""
        # Reference timestamps, one before 2000 one after 2000
        t2000_ref = (2001, 335, 53644260)
        t1900_ref = (1983, 336, 35058207)

        # Encoded version
        t2000_enc = np.array([847, 818, 35812])
        t1900_enc = np.array([42832, 534, 61983])

        # Test whether PODReader decodes them correctly
        self.assertEqual(GACPODReader.decode_timestamps(t2000_enc), t2000_ref,
                         msg="Timestamp after 2000 was decoded incorrectly")
        self.assertEqual(GACPODReader.decode_timestamps(t1900_enc), t1900_ref,
                         msg="Timestamp before 2000 was decoded incorrectly")

    @mock.patch("pygac.gac_pod.GACPODReader.decode_timestamps")
    def test_get_header_timestamp(self, decode_timestamps):
        """Test readout of header timestamp."""
        self.reader.head = {"start_time": 123}
        decode_timestamps.return_value = np.array(
            [2019]), np.array([123]), np.array([123456])
        time = self.reader.get_header_timestamp()
        decode_timestamps.assert_called_with(123)
        self.assertEqual(time, dt.datetime(2019, 5, 3, 0, 2, 3, 456000))

    @mock.patch("pygac.gac_pod.GACPODReader.decode_timestamps")
    def test_get_times(self, decode_timestamps):
        """Test getting times."""
        self.reader.scans = {"time_code": 123}
        self.reader._get_times_from_file()
        decode_timestamps.assert_called_with(123)

    def test_get_lonlat(self):
        """Test readout of lon/lat coordinates."""
        lons_exp = np.array([[2, 4],
                             [6, 8]])
        lats_exp = np.array([[1, 3],
                             [5, 7]])

        self.reader.scans = {"earth_location": {"lats": lats_exp * 128,
                                                "lons": lons_exp * 128}}

        lons, lats = self.reader._get_lonlat_from_file()
        numpy.testing.assert_array_equal(lons, lons_exp)
        numpy.testing.assert_array_equal(lats, lats_exp)

    @mock.patch("pygac.pod_reader.get_tsm_idx")
    def test_get_tsm_pixels(self, get_tsm_idx):
        """Test channel set used for TSM correction."""
        ones = np.ones((409, 100))
        zeros = np.zeros(ones.shape)
        ch1 = 1*ones
        ch2 = 2*ones
        ch4 = 4*ones
        ch5 = 5*ones
        channels = np.dstack((ch1, ch2, zeros, ch4, ch5))
        self.reader.get_tsm_pixels(channels)
        get_tsm_idx.assert_called_with(CalledWithArray(ch1),
                                       CalledWithArray(ch2),
                                       CalledWithArray(ch4),
                                       CalledWithArray(ch5))

    def test_quality_indicators(self):
        """Test the quality indicator unpacking."""
        reader = self.reader
        QFlag = reader.QFlag
        quality_indicators = np.array([
            1,  # 00...001
            QFlag.FATAL_FLAG,  # 100...00
            QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION,
            QFlag.TIME_ERROR | QFlag.DATA_GAP,
        ], dtype=">u4")
        # check if the bits look as expected
        bits = np.unpackbits(quality_indicators.view(np.uint8)).reshape((-1, 32))
        # For a big endian integer, the number 1 fills only the last of the 32 bits
        self.assertEqual(bits[0].sum(), 1)  # only one bit is filled
        self.assertEqual(bits[0][-1], 1)  # the last bit is filled
        # The fatal flag fills only the first bit
        self.assertEqual(bits[1].sum(), 1)  # only one bit is filled
        self.assertEqual(bits[1][0], 1)  # the first bit is filled

        # setup reader and test
        reader.scans = {self.reader._quality_indicators_key: quality_indicators}

        # default mask is QFlag.FATAL_FLAG | QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION
        expected_mask = np.array([False, True, True, False], dtype=bool)
        numpy.testing.assert_array_equal(reader.mask, expected_mask)

        # test individual flags
        expected_mask = np.array([False, False, False, True], dtype=bool)
        numpy.testing.assert_array_equal(
            reader._get_corrupt_mask(flags=QFlag.TIME_ERROR),
            expected_mask
        )
        # test combination of flags
        expected_mask = np.array([False, False, True, True], dtype=bool)
        flags = QFlag.DATA_GAP | QFlag.NO_EARTH_LOCATION
        numpy.testing.assert_array_equal(
            reader._get_corrupt_mask(flags=flags),
            expected_mask
        )

    @mock.patch("pygac.reader.get_lonlatalt")
    @mock.patch("pygac.reader.compute_pixels")
    @mock.patch("pygac.reader.Reader.get_tle_lines")
    @mock.patch("pygac.gac_reader.avhrr_gac_from_times")
    def test__adjust_clock_drift(self, avhrr_gac, get_tle_lines,
                                 compute_pixels, get_lonlatalt):
        """Test the clock drift adjustment."""
        sat_name = "fake_sat"
        reader = self.reader

        # We construct the following input
        # the scan lines do not start with 1 and have a gap
        scan_lines = np.array([15, 16, 17, 18, 22, 23, 24, 25])
        scan_rate = 0.5  # seconds/line
        # the first valid scans starts 1980-01-01 (without clock correction)
        # which leads to the following utcs for the given scan lines
        # ['1980-01-01T00:00:00.000', '1980-01-01T00:00:00.500',
        #  '1980-01-01T00:00:01.000', '1980-01-01T00:00:01.500',
        #  '1980-01-01T00:00:03.500', '1980-01-01T00:00:04.000',
        #  '1980-01-01T00:00:04.500', '1980-01-01T00:00:05.000']
        scan_utcs = (
            (1000 * scan_rate * (scan_lines - scan_lines[0])).astype("timedelta64[ms]")
            + np.datetime64("1980", "ms")
        )
        # For the geolocations, we assume an artificial swath of two pixels width
        # from north to south with constant lon, lat difference of 3deg for simplicity
        # lons = [[0, 3], [0, 3], [0, 3], [0, 3],
        #         [0, 3], [0, 3], [0, 3], [0, 3]],
        # lats = [[45, 45], [48, 48], [51, 51], [54, 54],
        #         [66, 66], [69, 69], [72, 72], [75, 75]]
        scan_angle = 3.0  # deg
        scan_lons, scan_lats = np.meshgrid(scan_angle*np.arange(2), scan_angle*scan_lines)

        # we assume a constant clock offset of 3.75 seconds
        # which should lead to the following adjustment on utcs
        # ['1979-12-31T23:59:56.250', '1979-12-31T23:59:56.750',
        #  '1979-12-31T23:59:57.250', '1979-12-31T23:59:57.750',
        #  '1979-12-31T23:59:59.750', '1980-01-01T00:00:00.250',
        #  '1980-01-01T00:00:00.750', '1980-01-01T00:00:01.250']
        offset = 3.75
        scan_offsets = offset*np.ones_like(scan_lines, dtype=float)  # seconds
        expected_utcs = scan_utcs - (1000*scan_offsets).astype("timedelta64[ms]")

        # the adjustment of geolocations should keep the lons unchanged,
        # but should shift the lats by scan_angel * offset / scan_rate
        # = 3deg/line * 3.75sec / 0.5sec/line = 22.5deg
        # [[22.5, 22.5], [25.5, 25.5], [28.5, 28.5], [31.5, 31.5],
        #  [43.5, 43.5], [46.5, 46.5], [49.5, 49.5], [52.5, 52.5]]
        expected_lons = scan_lons
        lats_shift = scan_angle*offset/scan_rate
        expected_lats = scan_lats - lats_shift

        # prepare the reader
        reader.scans = {"scan_line_number": scan_lines}
        reader._times_as_np_datetime64 = scan_utcs
        reader.lons = scan_lons
        reader.lats = scan_lats
        reader.spacecraft_name = sat_name

        # prepare offsets
        clock_offsets_txt[sat_name] = ("75001 000000 {offset}  "
                                       "85001 000000 {offset}").format(offset=offset)
        # set attitude coeffs
        reader._rpy = np.zeros(3)

        # set mocks for reader._compute_missing_lonlat call
        sgeom = mock.Mock()
        sgeom.times.return_value = [None]
        avhrr_gac.return_value = sgeom
        get_tle_lines.return_value = [None, None]
        compute_pixels.return_value = None
        # for the following mock, we need to know the missing values in advanced,
        # which we do, because we can convert the offset seconds into line number.
        offset_lines = offset / scan_rate
        min_line = np.floor(scan_lines[0] - offset_lines).astype(int)
        max_line = scan_lines[-1]
        missed_lines = np.setdiff1d(np.arange(min_line, max_line+1), scan_lines)
        n_missed = len(missed_lines)
        missed_lons = np.tile([0., scan_angle], n_missed)
        missed_lats = np.repeat(scan_angle*missed_lines, 2)
        get_lonlatalt.return_value = [missed_lons, missed_lats]

        # adjust clock drift
        reader.lonlat_sample_points = [0, 2]

        reader._adjust_clock_drift()

        # check output
        # use allclose for geolocations, because the slerp interpolation
        # includes a transormation to cartesian coordinates and back to lon, lats.
        numpy.testing.assert_array_equal(reader._times_as_np_datetime64, expected_utcs)
        numpy.testing.assert_allclose(reader.lons, expected_lons)
        numpy.testing.assert_allclose(reader.lats, expected_lats)

        # undo changes to clock_offsets_txt
        clock_offsets_txt.pop(sat_name)

    @mock.patch("pygac.pod_reader.get_offsets")
    @mock.patch("pygac.reader.Reader.get_tle_lines")
    def test__adjust_clock_drift_without_tle(self, get_tle_lines, get_offsets):
        """Test that clockdrift adjustment can handle missing TLE data."""
        reader = self.reader
        reader._times_as_np_datetime64 = np.zeros(10, dtype="datetime64[ms]")
        reader.scans = {"scan_line_number": np.arange(10)}
        get_offsets.return_value = np.zeros(10), np.zeros(10)
        get_tle_lines.side_effect = NoTLEData("No TLE data available")
        reader._adjust_clock_drift()  # should pass without errors


class TestPOD_truncate(unittest.TestCase):
    """Test the POD GAC reader."""

    longMessage = True

    def setUp(self):
        """Set up the test."""
        self.reader = GACPODReader()

    def test__two_logical_record(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # GAC POD should have two logical records per physical
        self.assertEqual(reader.offset // reader.scanline_type.itemsize, 2)

    def test__truncate_padding_record_even_correct(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # Even number of scans, correct file length
        reader.head = {"number_of_scans": 65534}
        buffer = reader._truncate_padding_record(bytes(65534*3220))
        self.assertEqual(len(buffer), 65534*3220)


    def test__truncate_padding_record_odd_correct(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # Odd number of scans, correct file length (should truncate)
        reader.head = {"number_of_scans": 65533}
        buffer = reader._truncate_padding_record(bytes(65534*3220))
        self.assertEqual(len(buffer), 65533*3220)


    def test__truncate_padding_record_odd_nopadding(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # Odd number of scans, padding record is missing
        with self.assertWarnsRegex(RuntimeWarning, "incomplete physical record"):
            reader.head = {"number_of_scans": 65533}
            buffer = reader._truncate_padding_record(bytes(65533*3220))
            self.assertEqual(len(buffer), 65533*3220)

    def test__truncate_padding_record_shortfile(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # File is too short (should do nothing)
        reader.head = {"number_of_scans": 65533}
        buffer = reader._truncate_padding_record(bytes(65532*3220))
        self.assertEqual(len(buffer), 65532*3220)

    def test__truncate_padding_record_longfile(self):
        """Test that truncate_padding_record is correctly dropping end padding"""
        reader = self.reader
        # File is too long (should do nothing)
        reader.head = {"number_of_scans": 65533}
        buffer = reader._truncate_padding_record(bytes(65535*3220))
        self.assertEqual(len(buffer), 65535*3220)



"""Convert PATMOS-x calibration file tarballs to PyGAC calibration json format.

The official tarballs are available on the PATMOS-x webpage "https://cimss.ssec.wisc.edu/patmosx/avhrr_cal.html".
"""

import argparse
import datetime as dt
import json
import logging
import pathlib
import re
import tarfile

from scipy.optimize import bisect


class PatmosxReader:
    """Read PATMOS-x coefficient files tarballs."""
    # regular expression with named capturing groups to read an entire patmosx file
    regex = re.compile(
        r"\s*(?P<sat_name>\w+)[^\n]*\n"
        r"\s*(?P<solar_3b>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ew_3b>[eE0-9\.-]+)[^\n]*\n"
        r'\s*(?P<ch3b_Ns>[eE0-9\.-]+),?\s*(?P<ch3b_b0>[eE0-9\.-]+),?\s*(?P<ch3b_b1>[eE0-9\.-]+),?\s*(?P<ch3b_b2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch4_Ns>[eE0-9\.-]+),?\s*(?P<ch4_b0>[eE0-9\.-]+),?\s*(?P<ch4_b1>[eE0-9\.-]+),?\s*(?P<ch4_b2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch5_Ns>[eE0-9\.-]+),?\s*(?P<ch5_b0>[eE0-9\.-]+),?\s*(?P<ch5_b1>[eE0-9\.-]+),?\s*(?P<ch5_b2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r"\s*(?P<nu_3b>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a1_3b>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a2_3b>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<nu_4>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a1_4>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a2_4>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<nu_5>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a1_5>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<a2_5>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch1_dark_count>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch2_dark_count>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch3a_dark_count>[eE0-9\.-]+)[^\n]*\n"
        r"(?:[a-z]+[^\n]*\n)?"
        r'\s*(?P<ch1_low_gain_S0>[eE0-9\.-]+)\s*(?P<ch1_low_gain_S1>[eE0-9\.-]+)\s*(?P<ch1_low_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch1_high_gain_S0>[eE0-9\.-]+)\s*(?P<ch1_high_gain_S1>[eE0-9\.-]+)\s*(?P<ch1_high_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch2_low_gain_S0>[eE0-9\.-]+)\s*(?P<ch2_low_gain_S1>[eE0-9\.-]+)\s*(?P<ch2_low_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch2_high_gain_S0>[eE0-9\.-]+)\s*(?P<ch2_high_gain_S1>[eE0-9\.-]+)\s*(?P<ch2_high_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch3a_low_gain_S0>[eE0-9\.-]+)\s*(?P<ch3a_low_gain_S1>[eE0-9\.-]+)\s*(?P<ch3a_low_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<ch3a_high_gain_S0>[eE0-9\.-]+)\s*(?P<ch3a_high_gain_S1>[eE0-9\.-]+)\s*(?P<ch3a_high_gain_S2>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r"\s*(?P<date_of_launch>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch1_gain_switches_count>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch2_gain_switches_count>[eE0-9\.-]+)[^\n]*\n"
        r"\s*(?P<ch3a_gain_switches_count>[eE0-9\.-]+)[^\n]*\n"
        r'\s*(?P<PRT1_0>[eE0-9\.-]+)\,*\s*(?P<PRT1_1>[eE0-9\.-]+)\,*\s*(?P<PRT1_2>[eE0-9\.-]+)\,*\s*(?P<PRT1_3>[eE0-9\.-]+)\,*\s*(?P<PRT1_4>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<PRT2_0>[eE0-9\.-]+)\,*\s*(?P<PRT2_1>[eE0-9\.-]+)\,*\s*(?P<PRT2_2>[eE0-9\.-]+)\,*\s*(?P<PRT2_3>[eE0-9\.-]+)\,*\s*(?P<PRT2_4>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<PRT3_0>[eE0-9\.-]+)\,*\s*(?P<PRT3_1>[eE0-9\.-]+)\,*\s*(?P<PRT3_2>[eE0-9\.-]+)\,*\s*(?P<PRT3_3>[eE0-9\.-]+)\,*\s*(?P<PRT3_4>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<PRT4_0>[eE0-9\.-]+)\,*\s*(?P<PRT4_1>[eE0-9\.-]+)\,*\s*(?P<PRT4_2>[eE0-9\.-]+)\,*\s*(?P<PRT4_3>[eE0-9\.-]+)\,*\s*(?P<PRT4_4>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'\s*(?P<PRT_weight_1>[eE0-9\.-]+)\s*(?P<PRT_weight_2>[eE0-9\.-]+)\s*(?P<PRT_weight_3>[eE0-9\.-]+)\s*(?P<PRT_weight_4>[eE0-9\.-]+)[^\n]*\n'  # noqa
        r'(?:\s*(?P<sst_mask_day_0>[eE0-9\.-]+),\s*(?P<sst_mask_day_1>[eE0-9\.-]+),\s*(?P<sst_mask_day_2>[eE0-9\.-]+),\s*(?P<sst_mask_day_3>[eE0-9\.-]+)[^\n]*\n)?'  # noqa
        r'(?:\s*(?P<sst_mask_night_0>[eE0-9\.-]+),\s*(?P<sst_mask_night_1>[eE0-9\.-]+),\s*(?P<sst_mask_night_2>[eE0-9\.-]+),\s*(?P<sst_mask_night_3>[eE0-9\.-]+)[^\n]*\n)?'  # noqa
        r"(?:\![^v][^\n]*\n)*"
        r"(?:\!(?P<version>v\w+))?"
    )

    def __init__(self, tarball):
        self.tarball = tarball
        self.coeffs = []
        with tarfile.open(tarball) as archive:
            for tarinfo in archive:
                if tarinfo.isfile():
                    # open satellite specific coefficient file
                    filename = tarinfo.name
                    fileobj = archive.extractfile(filename)
                    content = fileobj.read().decode()
                    match = self.regex.match(content)
                    sat_coeffs = {key: self.parse_types(value) for key, value in match.groupdict().items()}
                    self.coeffs.append(sat_coeffs)

    def __iter__(self):
        yield from self.coeffs  # noqa

    @staticmethod
    def parse_types(value):
        """parse the types of coefficients"""
        try:
            try:
                _value = int(value)
            except ValueError:
                _value = float(value)
        except ValueError:
            _value = value
        except TypeError:
            _value = None
        return _value


class Translator:
    """Translate PATMOS-x coefficients to PyGAC format."""
    sat_names = {"m01": "metopb", "m02": "metopa", "n05": "tirosn", "m03": "metopc"}
    sat_names.update({"n{0:02d}".format(i): "noaa{0}".format(i) for i in range(6,20)})
    description = {
        "visible": {
            "channels": ["1", "2", "3a"],
            "coefficients": {
                "dark_count": "instrument counts under dark conditions []",
                "gain_switch": "dual-gain switch count, set to 'null' for single-gain instruments []",
                "s0": "single-gain calibration slope at launch date [%]",
                "s1": "linear single-gain calibration slope parameter [% years^{-1}]",
                "s2": "quadratic single-gain calibration slope parameter [% years^{-2}]",
                "date_of_launch": "timestamp of launch date [UTC]"
            },
            "method": 'Heidinger, A.K., W.C. Straka III, C.C. Molling, J.T. Sullivan, and X. Wu, 2010: Deriving an inter-sensor consistent calibration for the AVHRR solar reflectance data record. International Journal of Remote Sensing, 31:6493-6517'  # noqa
        },
        "thermal": {
            "channels": ["3b", "4", "5"],
            "coefficients": {
                "centroid_wavenumber": "centroid wavenumber [cm^{-1}]",
                "b0": "constant non-linear radiance correction coefficient [mW m^{-2} sr cm^{-1}]",
                "b1": "linear non-linear radiance correction coefficient []",
                "b2": "quadratic non-linear radiance correction coefficient [(mW^{-1} m^2 sr^{-1} cm)]",
                "space_radiance": "radiance of space [mW m^{-2} sr cm^{-1}]",
                "to_eff_blackbody_intercept": "thermal channel temperature to effective blackbody temperature intercept [K]",  # noqa
                "to_eff_blackbody_slope": "thermal channel temperature to effective blackbody temperature slope []",
                "d0": "constant thermometer counts to temperature conversion coefficient [K]",
                "d1": "linear thermometer counts to temperature conversion coefficient [K]",
                "d2": "quadratic thermometer counts to temperature conversion coefficient [K]",
                "d3": "cubic thermometer counts to temperature conversion coefficient [K]",
                "d4": "quartic thermometer counts to temperature conversion coefficient [K]"
            },
            "method": "Goodrum, G., Kidwell, K.B. and W. Winston, 2000: NOAA KLM User's Guide. U.S. Department of Commerce, National Oceanic and Atmospheric Administration, National Environmental Satellite, Data and Information Service; Walton, C. C., J. T. Sullivan, C. R. N. Rao, and M. P. Weinreb, 1998: Corrections for detector nonlinearities and calibration inconsistencies of the infrared channels of the Advanced Very High Resolution Radiometer. J. Geophys. Res., 103, 3323-3337; Trishchenko, A.P., 2002: Removing Unwanted Fluctuations in the AVHRR Thermal Calibration Data Using Robust Techniques. Journal of Atmospheric and Oceanic Technology, 19:1939-1954"  # noqa
        }
    }

    def __init__(self, patmosx_coeffs):
        self.coeffs = {"description": self.description}
        for patmosx_sat_coeffs in patmosx_coeffs:
            sat_name = self.sat_names[patmosx_sat_coeffs["sat_name"]]
            pygac_sat_coeffs = self.convert(patmosx_sat_coeffs)
            self.coeffs[sat_name] = pygac_sat_coeffs

    @classmethod
    def convert(cls, patmosx_sat_coeffs):
        pygac_sat_coeffs = {}
        # visible calibration
        for ch in ("1", "2", "3a"):
            s0l = patmosx_sat_coeffs["ch{0}_low_gain_S0".format(ch)]
            s0h = patmosx_sat_coeffs["ch{0}_high_gain_S0".format(ch)]
            if s0l == s0h:
                gain_switch = None
                s0 = s0l
            else:
                gain_switch = patmosx_sat_coeffs["ch{0}_gain_switches_count".format(ch)]
                s0 = cls.find_s0(s0l, s0h, ch)
            pygac_sat_coeffs["channel_{0}".format(ch)] = {
                "dark_count": float(patmosx_sat_coeffs["ch{0}_dark_count".format(ch)]),
                "gain_switch": gain_switch,
                "s0": s0,
                "s1": patmosx_sat_coeffs["ch{0}_high_gain_S1".format(ch)],
                "s2": patmosx_sat_coeffs["ch{0}_high_gain_S2".format(ch)]
            }
        date_of_launch = cls.float2date(patmosx_sat_coeffs["date_of_launch"])
        pygac_sat_coeffs["date_of_launch"] = date_of_launch.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        # thermal channels
        for ch in ("3b", "4", "5"):
            pygac_sat_coeffs["channel_{0}".format(ch)] = {
                "b0": patmosx_sat_coeffs["ch{0}_b0".format(ch)],
                "b1": patmosx_sat_coeffs["ch{0}_b1".format(ch)],
                "b2": patmosx_sat_coeffs["ch{0}_b2".format(ch)],
                "centroid_wavenumber": patmosx_sat_coeffs["nu_{0}".format(ch)],
                "space_radiance": patmosx_sat_coeffs["ch{0}_Ns".format(ch)],
                "to_eff_blackbody_intercept": (-patmosx_sat_coeffs["a1_{0}".format(ch)]
                                                / patmosx_sat_coeffs["a2_{0}".format(ch)]),
                "to_eff_blackbody_slope": 1/patmosx_sat_coeffs["a2_{0}".format(ch)]
            }
        for t in range(1, 5):
            pygac_sat_coeffs["thermometer_{0}".format(t)] = {
                "d{0}".format(d): float(patmosx_sat_coeffs["PRT{0}_{1}".format(t, d)])
                for d in range(5)
            }
        return pygac_sat_coeffs

    @staticmethod
    def find_s0(s0_low, s0_high, ch):
        """Find the single-gain calibration slope at launch date.

        Arguments
            s0_low - low gain calibration slope at launch date
            s0_high - high gain calibration slope at launch date
            ch - channel name ("1", "2", "3a")

        Note:
            In case of a single-gain instrument, s0_low is equal to s0_high.
        """
        if s0_low == s0_high:
            # single gain case
            return s0_low
        if ch == "3a":
            g_low, g_high = 0.25, 1.75
        else:
            g_low, g_high = 0.5, 1.5

        # Note: the PATMOS-x coefficients are rounded to three decimals.
        def diff(s0): return s0_low - round(s0*g_low, 3) + s0_high - round(s0*g_high, 3)

        s0_l = s0_low / g_low
        s0_h = s0_high / g_high
        if diff(s0_l) == 0:
            s0 = s0_l
        elif diff(s0_h) == 0:
            s0 = s0_h
        else:
            s0 = bisect(diff, s0_l, s0_h)
        return s0

    @staticmethod
    def float2date(date_float):
        """Convert date float to date.

        Argument
            date_float (float) - date as year

        Return
            date (datetime.datetime) - date

        Note
            This is the reverse function of date2float.
        """
        year = int(date_float)
        days_in_year = (dt.datetime(year+1, 1, 1) - dt.datetime(year, 1, 1)).days
        seconds = date_float*days_in_year*24*3600 - year*days_in_year*24*3600
        diff = dt.timedelta(seconds=seconds)
        date = dt.datetime(year, 1, 1) + diff
        return date

    def save(self, filepath):
        """Save coefficients as PyGAC json file."""
        with open(filepath, mode="w") as json_file:
            json.dump(self.coeffs, json_file, indent=4, sort_keys=True)


def main():
    """The main function."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("tarball", type=str, help="path to PATMOS-x coefficients tarball")
    parser.add_argument("-o", "--output", type=str, metavar="JSON",
                        help='path to PyGAC json file, defaults to tarball path with suffix ".json"')
    parser.add_argument("-v", "--verbose", action="store_true", help="explain what is being done")
    args = parser.parse_args()
    if args.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING
    logging.basicConfig(level=loglevel, format="[%(asctime)s] %(message)s")
    tarball = pathlib.Path(args.tarball)
    logging.info('Read PATMOS-x tarball "%s".', tarball)
    patmosx_coeffs = PatmosxReader(tarball)
    logging.info("Translate PATMOS-x coefficients to PyGAC format.")
    pygac_coeffs = Translator(patmosx_coeffs)
    output = args.output or tarball.with_suffix(".json")
    logging.info('Write PyGAC calibration json file "%s".', output)
    pygac_coeffs.save(output)
    logging.info("Done!")

if __name__ == "__main__":
    main()



"""Test TSM module."""

import unittest

import numpy as np
import numpy.testing

import pygac.correct_tsm_issue as tsm


class TSMTest(unittest.TestCase):
    """Test TSM module."""

    def test_get_tsm_idx(self):
        """Test identification of TSM affected pixels."""
        cols = np.arange(409)
        rows = np.arange(3000)
        mcols, mrows = np.meshgrid(cols, rows)

        # Create dummy reflectances in [0, 1] and BTs [170, 350]
        refl = np.sin(mrows / float(rows.size)) * \
            np.cos(mcols / float(cols.size))
        bt = np.cos(mrows / float(rows.size)) * \
            np.sin(mcols / float(cols.size))
        bt = 170 + (350 - 170) * bt

        ch1 = refl.copy()
        ch2 = refl.copy()
        ch4 = bt.copy()
        ch5 = bt.copy()

        # Add rectangle with noise to channel 1 and 4
        noise_cols, noise_rows = np.meshgrid(np.arange(200, 300),
                                             np.arange(1000, 2000))
        rng = np.random.default_rng()
        for ch in [ch1, ch4]:
            ch[noise_rows, noise_cols] += 1000 + 1000*rng.random(size=noise_rows.shape)

        # We expect the filter to also detect the edges of the noisy
        # rectangle: At the edges there is a transition from zero to
        # nonzero channel diff which leads to an increased 3x3 standard
        # deviation.
        noise_cols_exp, noise_rows_exp = np.meshgrid(np.arange(199, 301),
                                                     np.arange(999, 2001))
        idx = tsm.get_tsm_idx(ch1, ch2, ch4, ch5)
        numpy.testing.assert_array_equal(idx[0], noise_rows_exp.ravel())
        numpy.testing.assert_array_equal(idx[1], noise_cols_exp.ravel())

    def test_std_filter(self):
        """Test standard deviation filter."""
        # Define test data
        data = np.array([
            [2., 2., 2., 2.],
            [2., np.nan, np.nan, 2.],
            [np.nan, 3., 3., np.nan],
            [3., 2., 2., 3.],
        ])

        # Define reference
        filtered_ref = np.sqrt(np.array([
            [0., 0., 0., 0.],
            [0.1875, 2./9., 2./9., 0.1875],
            [0.25, 0.25, 0.25, 0.25],
            [2./9., 0.24, 0.24, 2./9.]
        ]))

        # Apply filter and compare results against reference
        filtered = tsm.std_filter(data=data, box_size=3)
        numpy.testing.assert_array_equal(filtered, filtered_ref)



"""slerp implementation in numpy"""

import numpy as np


def tocart(lon, lat):
    rlat = np.deg2rad(lat)
    rlon = np.deg2rad(lon)
    x__ = np.cos(rlat) * np.cos(rlon)
    y__ = np.cos(rlat) * np.sin(rlon)
    z__ = np.sin(rlat)
    return np.dstack((x__, y__, z__))


def toll(arr):
    lat = np.arcsin(arr.take(2, -1))
    lon = np.arctan2(arr.take(1, -1), arr.take(0, -1))
    return np.rad2deg(np.dstack((lon.squeeze(), lat.squeeze())))


def dot(a, b):
    return np.sum(a * b, -1)


def slerp(lon0, lat0, lon1, lat1, t):
    cp0 = tocart(lon0, lat0)
    cp1 = tocart(lon1, lat1)

    dot_product = np.clip(dot(cp0, cp1), -1.0, 1.0)
    identical_mask = np.isclose(dot_product, 1.0)

    omega = np.arccos(dot_product)[:, :, np.newaxis]
    sin_omega = np.sin(omega)

    interp = (
        np.sin((1 - t) * omega) / np.where(sin_omega == 0, 1, sin_omega) * cp0
        + np.sin(t * omega) / np.where(sin_omega == 0, 1, sin_omega) * cp1
    )

    interp[identical_mask] = cp0[identical_mask]
    return toll(interp)



"""Generic reader for GAC and LAC data.

Can't be used as is, has to be subclassed to add specific read functions.
"""

import datetime
import logging
import os
import re
import types
import warnings
from abc import ABC, abstractmethod
from contextlib import suppress
from functools import cached_property
from importlib.metadata import entry_points

import geotiepoints as gtp
import numpy as np
import pyorbital
import xarray as xr
from packaging.version import Version
from pyorbital import astronomy
from pyorbital.geoloc import compute_pixels, get_lonlatalt
from pyorbital.orbital import Orbital

from pygac import gac_io
from pygac.utils import calculate_sun_earth_distance_correction, centered_modulus, get_absolute_azimuth_angle_diff

LOG = logging.getLogger(__name__)

# rpy values from
# here:http://yyy.rsmas.miami.edu/groups/rrsl/pathfinder/Processing/proc_app_a.html
rpy_coeffs = {
    "noaa7": {
        "roll": 0.000,
        "pitch": 0.000,
        "yaw": 0.000,
    },
    "noaa9": {
        "roll": 0.000,
        "pitch": 0.0025,
        "yaw": 0.000,
    },
    "noaa10": {
        "roll": 0.000,
        "pitch": 0.000,
        "yaw": 0.000,
    },
    "noaa11": {
        "roll": -0.0019,
        "pitch": -0.0037,
        "yaw": 0.000,
    },
    "noaa12": {
        "roll": 0.000,
        "pitch": 0.000,
        "yaw": 0.000,
    },
    "noaa14": {
        "roll": 0.000,
        "pitch": 0.000,
        "yaw": 0.000,
    },
}


class ReaderError(ValueError):
    """Raised in Reader.read if the given file does not correspond to it."""

    pass


class NoTLEData(IndexError):
    """Raised if no TLE data available within time range."""


class DecodingError(ValueError):
    """Raised when decoding of some value fails."""


class TimestampMismatch(IndexError):
    """Raise when matching timestamps doesn't work."""


class Reader(ABC):
    """Reader for GAC and LAC format, POD and KLM platforms."""

    # data set header format, see _validate_header for more details
    data_set_pattern = re.compile(r"\w{3}\.\w{4}\.\w{2}.D\d{5}\.S\d{4}\.E\d{4}\.B\d{7}\.\w{2}")

    def __init__(
        self,
        interpolate_coords=True,
        adjust_clock_drift=True,
        tle_dir=None,
        tle_name=None,
        tle_thresh=7,
        creation_site=None,
        custom_calibration=None,
        calibration_file=None,
        header_date="auto",
        calibration_method="noaa",
        calibration_parameters=None,
        correct_scanlines=True,
        reference_image=None,
        dem=None,
        compute_lonlats_from_tles: bool = False,
    ):
        """Init the reader.

        Args:
            interpolate_coords: Interpolate coordinates from tiepoint grid
                to all pixels.
            adjust_clock_drift: Adjust the geolocation to compensate for the
                clock error (POD satellites only).
            tle_dir: Directory holding TLE files
            tle_name: Filename pattern of TLE files.
            tle_thresh: Maximum number of days between observation and nearest
                TLE
            creation_site: The three-letter identifier of the creation site (eg 'NSS')
            custom_calibration: dictionary with a subset of user defined satellite specific
                                calibration coefficients
            calibration_file: path to json file containing default calibrations
            header_date: the date to use for pod header choice. Defaults to "auto".
            correct_scanlines: Remove corrrupt scanline numbers. Defaults to True
            reference_image: the reference image to use for georeferencing
            dem: the digital elevation model to use for orthocorrection
            compute_lonlats_from_tles: Do not use the longitudes and latitudes provided in the file, rather compute them
                                       from the TLE.

        """
        self.meta_data = {}
        self.interpolate_coords = interpolate_coords
        self.adjust_clock_drift = adjust_clock_drift
        self.tle_dir = tle_dir
        self.tle_name = tle_name
        self.tle_thresh = tle_thresh
        self.creation_site = (creation_site or "NSS").encode("utf-8")
        self.header_date = header_date
        self.correct_scanlines = correct_scanlines
        self.head = None
        self.scans = None
        self.spacecraft_name = None
        self.spacecraft_id = None
        self._times_as_np_datetime64 = None
        self.lats = None
        self.lons = None
        self.tle_lines = None
        self.filename = None
        self._mask = None
        self._rpy = None
        self.reference_image = reference_image
        self.dem = dem
        self.compute_lonlats_from_tles: bool = compute_lonlats_from_tles

        self.clock_drift_correction_applied = False

        if calibration_method.lower() not in ["noaa"]:
            raise ValueError(f"Unknown calibration method {calibration_method}.")
        self.calibration_method = calibration_method
        self.calibration_parameters = calibration_parameters or dict()
        if custom_calibration is not None or calibration_file is not None:
            warnings.warn(
                "`custom_calibration` and `calibration_file` parameters will be deprecated soon. "
                "Please use `calibration_parameters` dictionary instead.",
                PendingDeprecationWarning,
            )
            if not self.calibration_parameters:
                self.calibration_parameters = dict(custom_coeffs=custom_calibration, coeffs_file=calibration_file)

    @property
    def times(self):
        """Get the UTCs as datetime.datetime."""
        return self.to_datetime(self._times_as_np_datetime64)

    @property
    def filename(self):
        """Get the property 'filename'."""
        return self._filename

    @filename.setter
    def filename(self, filepath):
        """Set the property 'filename'."""
        if filepath is None:
            self._filename = None
        else:
            filepath = os.fspath(filepath)
            match = self.data_set_pattern.search(filepath)
            if match:
                self._filename = match.group()
            else:
                self._filename = os.path.basename(filepath)

    @abstractmethod
    def read(self, filename, fileobj=None):  # pragma: no cover
        """Read the GAC/LAC data.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def read_header(cls, filename, fileobj=None):  # pragma: no cover
        """Read the file header.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)

        Returns:
            archive_header (struct): archive header
            header (struct): file header

        Note:
            This is a classmethod to avoid throwaway instances while
            checking if the reader corresponds to the input file.
        """
        raise NotImplementedError

    @classmethod
    def _correct_data_set_name(cls, header, filename):
        """Replace invalid data_set_name from header with filename.

        Args:
            header (struct): file header
            filename (str): path to file
        """
        filename = str(filename)
        data_set_name = header["data_set_name"]
        try:
            header["data_set_name"] = cls._decode_data_set_name(data_set_name)
        except DecodingError:
            LOG.debug(f"The data_set_name in header {header['data_set_name']} does not match. Use filename instead.")
            match = cls.data_set_pattern.search(filename)
            if match:
                data_set_name = match.group()
                LOG.debug(f"Set data_set_name, to filename {data_set_name}")
                header["data_set_name"] = data_set_name.encode()
            else:
                LOG.debug(f"header['data_set_name']={header['data_set_name']}; filename='{filename}'")
                raise ReaderError("Cannot determine data_set_name!")
        return header

    @classmethod
    def _decode_data_set_name(cls, data_set_name):
        for encoding in "utf-8", "cp500":
            try:
                data_set_name = cls._decode_data_set_name_for_encoding(data_set_name, encoding)
            except DecodingError as err:
                LOG.debug(str(err))
            else:
                return data_set_name
        else:
            raise DecodingError("Could not reliably decode the dataset name.")

    @classmethod
    def _decode_data_set_name_for_encoding(cls, data_set_name, encoding):
        data_set_name = data_set_name.decode(encoding, errors="ignore")
        if not cls.data_set_pattern.match(data_set_name):
            raise DecodingError(
                f"The data_set_name in header {data_set_name} does not seem correct using encoding {encoding}."
            )
        else:
            data_set_name = data_set_name.encode()
        return data_set_name

    @classmethod
    def _validate_header(cls, header):
        """Check if the header belongs to this reader.

        Note:
            according to https://www1.ncdc.noaa.gov/pub/data/satellite/
            publications/podguides/TIROS-N%20thru%20N-14/pdf/NCDCPOD2.pdf
            and https://www1.ncdc.noaa.gov/pub/data/satellite/
            publications/podguides/N-15%20thru%20N-19/pdf/
            2.5%20Section%208.0%20NOAA%20Level%201B%20Database.pdf
            the data set name splits into
            PROCESSING-CENTER.DATA-TYPE.SPACECRAFT-UNIQUE-ID.
            YEAR-DAY.START-TIME.STOP-TIME.PROCESSING-BLOCK-ID.SOURCE
            This should be sufficient information to determine the
            reader.
            Global Area Coverage (GAC):
                DATA-TYPE = GHRR
            Local Area Coverage (LAC):
                DATA-TYPE = LHRR
            Polar Orbiter Data (POD):
                SPACECRAFT-UNIQUE-ID in [TN, NA, NB, NC, ND, NE, NF,
                                         NG, NH, NI, NJ]
            NOAA-K, -L, -M system, but also newer satellites (KLM):
                SPACECRAFT-UNIQUE-ID in [NK, NL, NM, NN, NP, M2, M1]
        """
        # This method does not need to be implemented in all subclasses.
        # It is intended for cooperative multiple inheritance, i.e.
        # each child class which implements this method, should call the
        # super method to enter into the method resolution order.
        # See https://docs.python.org/3/library/functions.html#super
        # second use case "diamond diagrams".
        # Check if the data set name matches the pattern
        LOG.debug("validate header")
        data_set_name = header["data_set_name"].decode(errors="ignore")
        if not cls.data_set_pattern.match(data_set_name):
            raise ReaderError("Data set name %s does not match!" % header["data_set_name"])

    def _read_scanlines(self, buffer, count):
        """Read the scanlines from the given buffer.

        Args:
            buffer (bytes, bytearray): buffer to read from
            count (int): number of expected scanlines
        """
        # Calculate the actual number of complete scanlines. The integer divisoin
        # may strip a potentially incomplete line at the end of the file.
        line_count = len(buffer) // self.scanline_type.itemsize
        if line_count != count:
            LOG.warning("Expected %d scan lines, but found %d!" % (count, line_count))
            warnings.warn("Unexpected number of scanlines!", category=RuntimeWarning, stacklevel=2)
        self.scans = np.frombuffer(buffer, dtype=self.scanline_type, count=line_count)

    @classmethod
    def can_read(cls, filename, fileobj=None):
        """Read the GAC/LAC data.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)

        Retruns:
            result (bool): True if the reader can read the input
        """
        if fileobj:
            pos = fileobj.tell()
        try:
            archive_header, header = cls.read_header(filename, fileobj=fileobj)
            result = True
        except (ReaderError, ValueError) as exception:
            LOG.debug("%s failed to read the file! %s" % (cls.__name__, repr(exception)))
            result = False
        finally:
            if fileobj:
                fileobj.seek(pos)
        return result

    @classmethod
    def fromfile(cls, filename, fileobj=None):
        """Create Reader from file, alternative constructor.

        Args:
            filename (str): Path to GAC/LAC file

        Kwargs:
            fileobj (file object): Open file object to read from

        Note:
            The fileobj is useful when dealing with tar archives,
            where the filename is given by the tarinfo object, but
            the extracted file object's property 'name' is set to
            the filename of the archive.
        """
        instance = cls()
        instance.read(filename, fileobj=fileobj)
        return instance

    def _get_calibrated_channels_uniform_shape(self):
        """Prepare the channels as input for gac_io.save_gac."""
        channels = self.get_calibrated_channels()
        assert channels.shape[-1] == 6
        return channels

    def save(
        self,
        start_line,
        end_line,
        output_file_prefix="PyGAC",
        output_dir="./",
        avhrr_dir=None,
        qual_dir=None,
        sunsatangles_dir=None,
    ):
        """Convert the Reader instance content into hdf5 files."""
        avhrr_dir = avhrr_dir or output_dir
        qual_dir = qual_dir or output_dir
        sunsatangles_dir = sunsatangles_dir or output_dir
        self.get_lonlat()
        channels = self._get_calibrated_channels_uniform_shape()
        sat_azi, sat_zen, sun_azi, sun_zen, rel_azi = self.get_angles()
        qual_flags = self.get_qual_flags()
        if np.all(self.mask):
            print("ERROR: All data is masked out. Stop processing")
            raise ValueError("All data is masked out.")
        gac_io.save_gac(
            self.spacecraft_name,
            self._times_as_np_datetime64,
            self.lats,
            self.lons,
            channels[:, :, 0],
            channels[:, :, 1],
            channels[:, :, 2],
            channels[:, :, 3],
            channels[:, :, 4],
            channels[:, :, 5],
            sun_zen,
            sat_zen,
            sun_azi,
            sat_azi,
            rel_azi,
            qual_flags,
            start_line,
            end_line,
            self.filename,
            self.meta_data,
            output_file_prefix,
            avhrr_dir,
            qual_dir,
            sunsatangles_dir,
        )

    @abstractmethod
    def get_header_timestamp(self):  # pragma: no cover
        """Read start timestamp from the header.

        Returns:
            datetime.datetime: Start timestamp

        """
        raise NotImplementedError

    def get_counts(self):
        """Get the counts.

        Returns:
            np.array: The counts, with channel 3a and 3b split if necessary.

        """
        packed_data = self.scans["sensor_data"]
        counts = np.zeros((len(self.scans), self.scan_width * 5))
        counts_nb = (self.scan_width * 5) // 3
        remainder = (self.scan_width * 5) % 3
        if remainder == 0:
            nb1 = nb2 = nb3 = counts_nb
        elif remainder == 1:
            nb1 = counts_nb + 1
            nb2 = nb3 = counts_nb
        elif remainder == 2:
            nb1 = nb2 = counts_nb + 1
            nb3 = counts_nb
        counts[:, 0::3] = ((packed_data >> 20) & 1023)[:, :nb1]
        counts[:, 1::3] = ((packed_data >> 10) & 1023)[:, :nb2]
        counts[:, 2::3] = (packed_data & 1023)[:, :nb3]
        counts = counts.reshape((-1, self.scan_width, 5))
        try:
            switch = self.get_ch3_switch()
        except AttributeError:
            return counts
        else:
            channels = np.zeros((len(self.scans), self.scan_width, 6), dtype=counts.dtype)
            channels[:, :, :2] = counts[:, :, :2]
            channels[:, :, -2:] = counts[:, :, -2:]
            channels[:, :, 2][switch == 1] = counts[:, :, 2][switch == 1]
            channels[:, :, 3][switch == 0] = counts[:, :, 2][switch == 0]
            return channels

    @abstractmethod
    def _get_times_from_file(self):  # pragma: no cover
        """Specify how to read scanline timestamps from data.

        Returns:
            int: year
            int: day of year
            int: milliseconds since 00:00

        """
        raise NotImplementedError

    def get_times(self):
        """Read scanline timestamps and try to correct invalid values.

        Returns:
            UTC timestamps

        """
        if self._times_as_np_datetime64 is None:
            # Read timestamps
            year, jday, msec = self._get_times_from_file()

            # Correct invalid values
            year, jday, msec = self.correct_times_median(year=year, jday=jday, msec=msec)
            self._times_as_np_datetime64 = self.to_datetime64(year=year, jday=jday, msec=msec)
            try:
                self._times_as_np_datetime64 = self.correct_times_thresh()
            except TimestampMismatch as err:
                LOG.error(str(err))

        return self._times_as_np_datetime64

    @staticmethod
    def to_datetime64(year, jday, msec):
        """Convert timestamps to numpy.datetime64.

        Args:
            year: Year
            jday: Day of the year (1-based)
            msec: Milliseconds since 00:00

        Returns:
            numpy.datetime64: Converted timestamps

        """
        return (
            year.astype(str).astype("datetime64[Y]")
            + (jday - 1).astype("timedelta64[D]")
            + msec.astype("timedelta64[ms]")
        )

    @staticmethod
    def to_datetime(datetime64):
        """Convert numpy.datetime64 to datetime.datetime.

        Args:
            datetime64 (numpy.datetime64): Numpy timestamp to be converted.

        Returns:
            datetime.datetime: Converted timestamp

        """
        return datetime64.astype(datetime.datetime)

    def lineno2msec(self, scan_line_number):
        """Compute ideal scanline timestamp based on the scanline number.

        Assumes a constant scanning frequency.

        Args:
            scan_line_number: Specifies the scanline number (1-based)

        Returns:
            Corresponding timestamps in milliseconds since 1970-01-01 00:00,
            i.e. the first scanline has timestamp 0.

        """
        return (scan_line_number - 1) / self.scan_freq

    def get_sun_earth_distance_correction(self):
        """Get the julian day and the sun-earth distance correction."""
        self.get_times()
        # Shouldn't we use the start time from the header if available?
        start_time = self._times_as_np_datetime64[0]
        jday = (start_time - start_time.astype("datetime64[Y]")).astype("timedelta64[D]").astype(int) + 1
        return calculate_sun_earth_distance_correction(jday)

    def update_meta_data(self):
        """Add some meta data to the meta_data dicitonary."""
        meta_data = self.meta_data
        self._update_meta_data_object(meta_data)
        if "gac_header" not in meta_data:
            meta_data["gac_header"] = self.head.copy()

    def _update_meta_data_object(self, meta_data):
        if "sun_earth_distance_correction_factor" not in meta_data:
            meta_data["sun_earth_distance_correction_factor"] = self.get_sun_earth_distance_correction()
        if "midnight_scanline" not in meta_data:
            meta_data["midnight_scanline"] = self.get_midnight_scanline()
        if "missing_scanlines" not in meta_data:
            meta_data["missing_scanlines"] = self.get_miss_lines()

    def read_as_dataset(self, file_to_read):
        self.read(file_to_read)
        return self.create_counts_dataset()

    def create_counts_dataset(self):
        head = dict(zip(self.head.dtype.names, self.head.item(), strict=False))
        scans = self.scans

        counts = self.get_counts()
        line_numbers = scans["scan_line_number"]
        scan_size = counts.shape[1]
        columns = np.arange(scan_size)
        longitudes, latitudes = self._get_lonlat_dataarrays(line_numbers, columns)

        times = self.get_times()

        if counts.shape[-1] == 5:
            channel_names = ["1", "2", "3", "4", "5"]
            ir_channel_names = ["3", "4", "5"]
        else:
            channel_names = ["1", "2", "3a", "3b", "4", "5"]
            ir_channel_names = ["3b", "4", "5"]

        channels = xr.DataArray(
            counts,
            dims=["scan_line_index", "columns", "channel_name"],
            coords=dict(
                scan_line_index=line_numbers,
                channel_name=channel_names,
                columns=columns,
                times=("scan_line_index", times),
            ),
        )

        prt, ict, space = self._get_telemetry_dataarrays(line_numbers, ir_channel_names)

        if self.interpolate_coords:
            channels = channels.assign_coords(
                longitude=(("scan_line_index", "columns"), longitudes.reindex_like(channels).data),
                latitude=(("scan_line_index", "columns"), latitudes.reindex_like(channels).data),
            )

        ds = xr.Dataset(
            dict(
                channels=channels,
                prt_counts=prt,
                ict_counts=ict,
                space_counts=space,
                longitude=longitudes,
                latitude=latitudes,
            ),
            attrs=head,
        )

        ds.attrs["spacecraft_name"] = self.spacecraft_name
        ds.attrs["max_scan_angle"] = 55.25 if self.spacecraft_name == "noaa16" else 55.37
        self._update_meta_data_object(ds.attrs)
        return ds

    def _get_lonlat_dataarrays(self, line_numbers, columns):
        lons, lats = self.get_lonlat()
        lon_lat_columns = columns if self.interpolate_coords is True else self.lonlat_sample_points

        if self.interpolate_coords:
            longitudes = xr.DataArray(
                lons,
                dims=["scan_line_index", "columns"],
                coords={"scan_line_index": line_numbers, "columns": lon_lat_columns},
            )
            latitudes = xr.DataArray(
                lats,
                dims=["scan_line_index", "columns"],
                coords={"scan_line_index": line_numbers, "columns": lon_lat_columns},
            )

        else:
            longitudes = xr.DataArray(
                lons,
                dims=["scan_line_index", "subsampled_columns"],
                coords={"scan_line_index": line_numbers, "subsampled_columns": lon_lat_columns},
            )
            latitudes = xr.DataArray(
                lats,
                dims=["scan_line_index", "subsampled_columns"],
                coords={"scan_line_index": line_numbers, "subsampled_columns": lon_lat_columns},
            )

        return longitudes, latitudes

    def _get_telemetry_dataarrays(self, line_numbers, ir_channel_names):
        prt, ict, space = self.get_telemetry()

        prt = xr.DataArray(prt, dims=["scan_line_index"], coords=dict(scan_line_index=line_numbers))
        ict = xr.DataArray(
            ict,
            dims=["scan_line_index", "ir_channel_name"],
            coords=dict(ir_channel_name=ir_channel_names, scan_line_index=line_numbers),
        )
        space = xr.DataArray(
            space,
            dims=["scan_line_index", "ir_channel_name"],
            coords=dict(ir_channel_name=ir_channel_names, scan_line_index=line_numbers),
        )

        return prt, ict, space

    def get_calibrated_channels(self):
        """Calibrate and return the channels."""
        calibrated_ds = self.calibrated_dataset

        channels = calibrated_ds["channels"].data
        with suppress(KeyError):
            self.meta_data["calib_coeffs_version"] = calibrated_ds.attrs["calib_coeffs_version"]

        return channels

    @cached_property
    def calibrated_dataset(self):
        """Return the calibrated dataset."""
        return self.get_calibrated_dataset()

    def get_calibrated_dataset(self):
        """Create and calibrate the dataset for the pass."""
        ds = self.create_counts_dataset()
        # calibration = {"1": "mitram", "2": "mitram", "4": {"method":"noaa", "coeff_file": "myfile.json"}}

        calibration_entrypoints = entry_points(group="pygac.calibration")
        calibration_function = calibration_entrypoints[self.calibration_method].load()
        calibrated_ds = calibration_function(ds, **self.calibration_parameters)

        # Mask out corrupt values
        mask = xr.DataArray(self.mask == False, dims=["scan_line_index"])  # noqa
        calibrated_ds = calibrated_ds.where(mask)

        # Apply KLM/POD specific postprocessing
        self.postproc(calibrated_ds)

        calibrated_ds.attrs["tle"] = self.get_tle_lines()

        # Mask pixels affected by scan motor issue
        if self.is_tsm_affected():
            LOG.info("Correcting for temporary scan motor issue")
            self.mask_tsm_pixels(calibrated_ds)
        if self.reference_image:
            self._georeference_data(calibrated_ds)
        return calibrated_ds

    def _georeference_data(self, calibrated_ds):
        if not self.adjust_clock_drift:
            self._correct_time_offset(calibrated_ds)

        from georeferencer.georeferencer import get_swath_displacement

        _, sat_zen, _, sun_zen, _ = self.get_angles()
        time_diff, (roll, pitch, yaw), (odistances, mdistances) = get_swath_displacement(
            calibrated_ds, sun_zen, sat_zen, self.reference_image, self.dem
        )

        if mdist := np.median(mdistances) > 5000:
            raise RuntimeError("Displacement minimization did not produce convincing improvements")
        calibrated_ds.attrs["median_gcp_distance"] = mdist

        self._rpy = roll, pitch, yaw
        time_diff = np.timedelta64(int(time_diff * 1e9), "ns")
        lons, lats = self._compute_lonlats(time_offset=time_diff)
        self._times_as_np_datetime64 += time_diff
        calibrated_ds["longitude"].data = lons
        calibrated_ds["latitude"].data = lats
        if self.dem:
            from georeferencer.georeferencer import orthocorrection

            calibrated_ds = orthocorrection(calibrated_ds, sat_zen, self.dem)
        calibrated_ds["times"].data = self._times_as_np_datetime64

    def _correct_time_offset(self, calibrated_ds) -> None:
        thinned_lons, thinned_lats = self._get_lonlat_from_file()
        gcps = np.array(
            (
                (0, self.lonlat_sample_points[0]),
                (0, self.lonlat_sample_points[-1]),
                (len(self.scans) - 1, self.lonlat_sample_points[0]),
                (len(self.scans) - 1, self.lonlat_sample_points[-1]),
            )
        )
        ref_lons = (thinned_lons[0, 0], thinned_lons[0, -1], thinned_lons[-1, 0], thinned_lons[-1, -1])
        ref_lats = (thinned_lats[0, 0], thinned_lats[0, -1], thinned_lats[-1, 0], thinned_lats[-1, -1])
        from pyorbital.geoloc_avhrr import estimate_time_offset

        time_diff, _ = estimate_time_offset(
            gcps,
            ref_lons,
            ref_lats,
            calibrated_ds["times"][0].values,
            calibrated_ds.attrs["tle"],
            calibrated_ds.attrs["max_scan_angle"],
        )
        time_diff = np.timedelta64(int(time_diff * 1e9), "ns")
        lons, lats = self._compute_lonlats(time_offset=time_diff)
        self._times_as_np_datetime64 += time_diff

        calibrated_ds["longitude"].data = lons
        calibrated_ds["latitude"].data = lats
        calibrated_ds["times"].data = self._times_as_np_datetime64

    @abstractmethod
    def get_telemetry(self):  # pragma: no cover
        """KLM/POD specific readout of telemetry."""
        raise NotImplementedError

    def lonlat_interpolator(self, lons, lats, cols_subset=None, cols_full=None):
        """Interpolate from lat-lon tie-points to pixel locations

        Args:
            lons: Longitude tie-points
            lats: Latitude tie-points

        Returns:
            pixel_longitudes, pixel_latitudes
        """
        if cols_subset is None and cols_full is None:
            cols_subset = self.lonlat_sample_points
            cols_full = np.arange(self.scan_width)
        rows = np.arange(len(lats))

        along_track_order = 1
        cross_track_order = 3

        satint = gtp.SatelliteInterpolator(
            (lons, lats), (rows, cols_subset), (rows, cols_full), along_track_order, cross_track_order
        )

        return satint.interpolate()

    def get_lonlat(self):
        """Compute lat/lon coordinates.

        TODO: Switch to faster interpolator?
        """
        if self.lons is None or self.lats is None:
            return self._compute_lonlats(mask_scanlines=not self.reference_image)

        return self.lons, self.lats

    def _compute_lonlats(self, time_offset=None, mask_scanlines=True):
        if not self.compute_lonlats_from_tles:
            self.lons, self.lats = self._get_lonlat_from_file()
            # Adjust clock drift
            if self.adjust_clock_drift and not self.clock_drift_correction_applied:
                self._adjust_clock_drift()
                self.clock_drift_correction_applied = True
        else:
            self.get_times()
            if self.adjust_clock_drift and not self.clock_drift_correction_applied:
                try:
                    offsets = self.compute_clock_offsets()
                    self._times_as_np_datetime64 -= (offsets * 1000).astype("timedelta64[ms]")
                    self.clock_drift_correction_applied = True
                    LOG.debug("Applied clock drift correction")
                except AttributeError:
                    LOG.debug("No clock drift correction applied")
            if time_offset:
                new_times = self._times_as_np_datetime64 + time_offset
                self.lons, self.lats = self._compute_lonlat_from_tles(new_times.astype("datetime64[ms]"))
            else:
                self.lons, self.lats = self._compute_lonlat_from_tles(self._times_as_np_datetime64)
        self.update_meta_data()

        # Interpolate from every eighth pixel to all pixels.
        if self.interpolate_coords:
            self.lons, self.lats = self.lonlat_interpolator(self.lons, self.lats)

        # Mask out corrupt scanlines
        if mask_scanlines:
            self.lons[self.mask] = np.nan
            self.lats[self.mask] = np.nan

        # Mask values outside the valid range
        self.lats[np.fabs(self.lats) > 90.0] = np.nan
        self.lons[np.fabs(self.lons) > 180.0] = np.nan

        return self.lons, self.lats

    @abstractmethod
    def _get_lonlat_from_file(self):  # pragma: no cover
        """KLM/POD specific readout of lat/lon coordinates."""
        raise NotImplementedError

    @property
    def mask(self):
        """Mask for corrupt scanlines."""
        if self._mask is None:
            self._mask = self._get_corrupt_mask()
        return self._mask

    @property
    @abstractmethod
    def QFlag(self):  # pragma: no cover
        """KLM/POD specific quality indicators."""
        raise NotImplementedError

    @property
    @abstractmethod
    def _quality_indicators_key(self):  # pragma: no cover
        raise NotImplementedError

    def _get_corrupt_mask(self, flags=None):
        """Readout of corrupt scanline mask.

        Args:
            flags (QFlag.flag): An "ORed" bitmask that defines corrupt values.
                                Defauts to (QFlag.FATAL_FLAG | QFlag.CALIBRATION
                                | QFlag.NO_EARTH_LOCATION)

        Note:
            The Quality flags mapping (QFlag) is KLM/POD specific.
        """
        QFlag = self.QFlag
        if flags is None:
            flags = QFlag.FATAL_FLAG | QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION
        return (self.scans[self._quality_indicators_key] & int(flags)).astype(bool)

    def get_qual_flags(self):
        """Read quality flags."""
        number_of_scans = self.scans["telemetry"].shape[0]
        qual_flags = np.zeros((int(number_of_scans), 7))
        qual_flags[:, 0] = self.scans["scan_line_number"]
        qual_flags[:, 1] = self._get_corrupt_mask(flags=self.QFlag.FATAL_FLAG)
        qual_flags[:, 2] = self._get_corrupt_mask(flags=self.QFlag.CALIBRATION)
        qual_flags[:, 3] = self._get_corrupt_mask(flags=self.QFlag.NO_EARTH_LOCATION)
        qual_flags[:, 4] = self._get_corrupt_mask(flags=self.QFlag.CH_3_CONTAMINATION)
        qual_flags[:, 5] = self._get_corrupt_mask(flags=self.QFlag.CH_4_CONTAMINATION)
        qual_flags[:, 6] = self._get_corrupt_mask(flags=self.QFlag.CH_5_CONTAMINATION)
        return qual_flags

    @abstractmethod
    def postproc(self, ds):  # pragma: no cover
        """Apply KLM/POD specific postprocessing."""
        raise NotImplementedError

    @abstractmethod
    def _adjust_clock_drift(self):  # pragma: no cover
        """Adjust clock drift."""
        raise NotImplementedError

    @staticmethod
    def tle2datetime64(times):
        """Convert TLE timestamps to numpy.datetime64.

        Args:
           times (float): TLE timestamps as %y%j.1234, e.g. 18001.25
        """
        # Convert %y%j.12345 to %Y%j.12345 (valid for 1950-2049)
        times = np.where(times > 50000, times + 1900000, times + 2000000)

        # Convert float to datetime64
        doys = (times % 1000).astype("int") - 1
        years = (times // 1000).astype("int")
        msecs = np.rint(24 * 3600 * 1000 * (times % 1))
        times64 = (years - 1970).astype("datetime64[Y]").astype("datetime64[ms]")
        times64 += doys.astype("timedelta64[D]")
        times64 += msecs.astype("timedelta64[ms]")

        return times64

    def get_tle_file(self):
        """Find TLE file for the current satellite."""
        tle_dir, tle_name = self.tle_dir, self.tle_name
        if tle_dir is None:
            raise RuntimeError("TLE directory not specified!")
        if tle_name is None:
            raise RuntimeError("TLE name not specified!")
        values = {
            "satname": self.spacecraft_name,
        }
        tle_filename = os.path.join(tle_dir, tle_name % values)
        LOG.info("TLE filename = " + str(tle_filename))

        return tle_filename

    def read_tle_file(self, tle_filename):
        """Read TLE file."""
        with open(tle_filename, "r") as fp_:
            return [line for line in fp_.readlines() if line.strip()]

    def get_tle_lines(self):
        """Find closest two line elements (TLEs) for the current orbit.

        Raises:
            IndexError, if the closest TLE is more than :meth:`pygac.GACReader.tle_thresh` days apart

        """
        if self.tle_lines is not None:
            return self.tle_lines
        self.get_times()
        tle_data = self.read_tle_file(self.get_tle_file())
        sdate = self._times_as_np_datetime64[0]
        dates = self.tle2datetime64(np.array([float(line[18:32]) for line in tle_data[::2]]))

        # Find index "iindex" such that dates[iindex-1] < sdate <= dates[iindex]
        # Notes:
        #     1. If sdate < dates[0] then iindex = 0
        #     2. If sdate > dates[-1] then iindex = len(dates), beyond the right boundary!
        iindex = np.searchsorted(dates, sdate)

        if iindex in (0, len(dates)):
            if iindex == len(dates):
                # Reset index if beyond the right boundary (see note 2. above)
                iindex -= 1
        elif abs(sdate - dates[iindex - 1]) < abs(sdate - dates[iindex]):
            # Choose the closest of the two surrounding dates
            iindex -= 1

        # Make sure the TLE we found is within the threshold
        delta_days = abs(sdate - dates[iindex]) / np.timedelta64(1, "D")
        if delta_days > self.tle_thresh:
            raise NoTLEData(
                "Can't find tle data for %s within +/- %d days around %s"
                % (self.spacecraft_name, self.tle_thresh, sdate)
            )

        if delta_days > 3:
            LOG.warning("Found TLE data for %s that is %f days appart", sdate, delta_days)
        else:
            LOG.debug("Found TLE data for %s that is %f days appart", sdate, delta_days)

        # Select TLE data
        tle1 = tle_data[iindex * 2]
        tle2 = tle_data[iindex * 2 + 1]
        self.tle_lines = tle1, tle2
        return tle1, tle2

    def get_sat_angles(self):
        """Get satellite angles.

        Returns:
            Azimuth, elevation (degrees)
        """
        try:
            return self._get_sat_angles_with_tle()
        except NoTLEData:
            LOG.warning("No TLE data available. Falling back to approximate calculation of satellite angles.")
            return self._get_sat_angles_without_tle()

    def _get_sat_angles_with_tle(self):
        tle1, tle2 = self.get_tle_lines()
        orb = Orbital(self.spacecrafts_orbital[self.spacecraft_id], line1=tle1, line2=tle2)
        sat_azi, sat_elev = orb.get_observer_look(self._times_as_np_datetime64[:, np.newaxis], self.lons, self.lats, 0)
        return sat_azi, sat_elev

    def _get_sat_angles_without_tle(self):
        """Get satellite angles using lat/lon from data to approximate satellite postition instead of TLE."""
        from pyorbital.orbital import get_observer_look as get_observer_look_no_tle

        LOG.warning("Approximating satellite height to 850km (TIROS-N OSCAR)!")
        sat_alt = 850.0  # km  TIROS-N OSCAR
        mid_column = int(0.5 * self.lons.shape[1])
        sat_azi, sat_elev = get_observer_look_no_tle(
            self.lons[:, mid_column][:, np.newaxis],
            self.lats[:, mid_column][:, np.newaxis],  # approximate satellite position
            sat_alt,  # approximate satellite altitude
            self._times_as_np_datetime64[:, np.newaxis],
            self.lons,
            self.lats,
            0,
        )
        # Sometimes (pyorbital <= 1.6.1) the get_observer_look_not_tle returns nodata instead of 90.
        # Problem solved with https://github.com/pytroll/pyorbital/pull/77
        if Version(pyorbital.__version__) <= Version("1.6.1"):
            sat_elev[:, mid_column] = 90
        return sat_azi, sat_elev

    def get_angles(self):
        """Get azimuth and zenith angles.

        Azimuth angle definition is the same as in pyorbital, but with
        different units (degrees not radians for sun azimuth angles)
        and different ranges.

        Returns:
            sat_azi: satellite azimuth angle degree clockwise from north in
            range ]-180, 180]

            sat_zenith: satellite zenith angles in degrees in range [0,90]

            sun_azi: sun azimuth angle degree clockwise from north in range
            ]-180, 180]

            sun_zenith: sun zenith angles in degrees in range [0,90]

            rel_azi: absolute azimuth angle difference in degrees between sun
            and sensor in range [0, 180]

        """
        self.get_times()
        self.get_lonlat()
        times = self._times_as_np_datetime64
        sat_azi, sat_elev = self.get_sat_angles()

        sat_zenith = 90 - sat_elev
        sun_zenith = astronomy.sun_zenith_angle(times[:, np.newaxis], self.lons, self.lats)

        alt, sun_azi = astronomy.get_alt_az(times[:, np.newaxis], self.lons, self.lats)
        del alt
        sun_azi = np.rad2deg(sun_azi)
        rel_azi = get_absolute_azimuth_angle_diff(sun_azi, sat_azi)

        # Scale angles range to half open interval ]-180, 180]
        sat_azi = centered_modulus(sat_azi, 360.0)
        sun_azi = centered_modulus(sun_azi, 360.0)

        # Mask corrupt scanlines
        for arr in (sat_azi, sat_zenith, sun_azi, sun_zenith, rel_azi):
            arr[self.mask] = np.nan

        return sat_azi, sat_zenith, sun_azi, sun_zenith, rel_azi

    def correct_times_median(self, year, jday, msec):
        """Replace invalid timestamps with statistical estimates (using median).

        Assumes that the majority of timestamps is ok.

        Args:
            year: Year
            jday: Day of the year
            msec: Milliseconds since 00:00

        Returns:
            Corrected year
            Corrected day of the year
            Corrected milliseconds

        """
        # Estimate ideal timestamps based on the scanline number. Still without
        # offset, e.g. the first scanline has timestamp 1970-01-01 00:00
        msec_lineno = self.lineno2msec(self.scans["scan_line_number"])

        jday = np.where(np.logical_or(jday < 1, jday > 366), np.median(jday), jday)
        if_wrong_jday = np.ediff1d(jday, to_begin=jday.dtype.type(0))
        jday = np.where(if_wrong_jday < 0, max(jday), jday)

        if_wrong_msec = np.where(msec < 1)
        if_wrong_msec = if_wrong_msec[0]
        if len(if_wrong_msec) > 0:
            if if_wrong_msec[0] != 0:
                msec = msec[0] + msec_lineno
            else:
                msec0 = np.median(msec - msec_lineno)
                msec = msec0 + msec_lineno

        if_wrong_msec = np.ediff1d(msec, to_begin=msec.dtype.type(0))
        msec = np.where(
            np.logical_and(np.logical_or(if_wrong_msec < -1000, if_wrong_msec > 1000), if_wrong_jday != 1),
            msec[0] + msec_lineno,
            msec,
        )

        # checking if year value is out of valid range
        if_wrong_year = np.where(np.logical_or(year < 1978, year > datetime.datetime.now().year))
        if_wrong_year = if_wrong_year[0]
        if len(if_wrong_year) > 0:
            # if the first scanline has valid time stamp
            if if_wrong_year[0] != 0:
                year = year[0]
                jday = jday[0]
                msec = msec[0] + msec_lineno
            # Otherwise use median time stamp
            else:
                year = np.median(year)
                jday = np.median(jday)
                msec0 = np.median(msec - msec_lineno)
                msec = msec0 + msec_lineno

        return year.astype(int), jday.astype(int), msec

    def correct_scan_line_numbers(self):
        """Remove scanlines with corrupted scanline numbers.

        This includes:
            - Scanline numbers outside the valid range
            - Scanline numbers deviating more than a certain threshold from the
              ideal case (1,2,3,...N)

        Example files having corrupt scanline numbers:
            - NSS.GHRR.NJ.D96144.S2000.E2148.B0720102.GC
            - NSS.GHRR.NJ.D96064.S0043.E0236.B0606162.WI
            - NSS.GHRR.NJ.D99286.S1818.E2001.B2466869.WI

        Returns:
            Intermediate and final results (for plotting purpose)

        """
        along_track = np.arange(1, len(self.scans["scan_line_number"]) + 1)
        results = {"along_track": along_track, "n_orig": self.scans["scan_line_number"].copy()}

        # Remove scanlines whose scanline number is outside the valid range
        within_range = np.logical_and(
            self.scans["scan_line_number"] < self.max_scanlines, self.scans["scan_line_number"] >= 0
        )
        self.scans = self.scans[within_range]

        # Remove scanlines deviating more than a certain threshold from the
        # ideal case (1,2,3,...N).
        ideal = np.arange(1, len(self.scans["scan_line_number"]) + 1)

        # ... Estimate possible offset (in case some scanlines are missing in
        # the beginning of the scan)
        offsets = self.scans["scan_line_number"] - ideal
        med_offset = np.median(offsets)

        # ... Compute difference to ideal case (1,2,3,...N) + offset
        diffs = np.abs(self.scans["scan_line_number"] - (ideal + med_offset))

        # ... Remove those scanlines whose difference is larger than a certain
        # threshold. For the threshold computation we only regard nonzero
        # differences.
        nz_diffs = diffs[diffs > 0]
        if len(nz_diffs) < 50:
            # Not enough differences for reliable statistics. Use fixed
            # threshold.
            thresh = 500
        else:
            mean_nz_diffs = np.mean(nz_diffs)
            std_nz_diffs = np.std(nz_diffs)
            med_nz_diffs = np.median(nz_diffs)
            mad_nz_diffs = np.median(np.abs(nz_diffs - med_nz_diffs))
            if mean_nz_diffs / float(med_nz_diffs) < 3:
                # Relatively small variation, keep (almost) everything
                thresh = mean_nz_diffs + 3 * std_nz_diffs
            else:
                # Large variation, filter more aggressively. Use median and
                # median absolute deviation (MAD) as they are less sensitive to
                # outliers. However, allow differences < 500 scanlines as they
                # occur quite often.
                thresh = max(500, med_nz_diffs + 3 * mad_nz_diffs)
        self.scans = self.scans[diffs <= thresh]

        LOG.debug("Removed %s scanline(s) with corrupt scanline numbers", str(len(along_track) - len(self.scans)))

        results.update(
            {
                "n_corr": self.scans["scan_line_number"],
                "within_range": within_range,
                "diffs": diffs,
                "thresh": thresh,
                "nz_diffs": nz_diffs,
            }
        )
        return results

    def correct_times_thresh(
        self, max_diff_from_t0_head=6 * 60 * 1000, min_frac_near_t0_head=0.01, max_diff_from_ideal_t=10 * 1000
    ):
        """Correct corrupted timestamps using a threshold approach.

        The threshold approach is based on the scanline number and the header
        timestamp. It also works if the majority of scanlines has corrupted
        timestamps.

        The header timestamp is used as a guideline to estimate the offset
        between timestamps computed from the scanline number and the actual
        scanline timestamps in the data. If header timestamp and scanline
        timestamps do not match, no correction is applied.

        Once the offset has been estimated, one can calculate the ideal
        timestamps based on the scanline number. Timestamps deviating more than
        a certain threshold from the ideal timestamps are replaced by
        the ideal timestamps.

        Example files having corrupt timestamps:
            - NSS.GHRR.NA.D81193.S2329.E0116.B1061214.WI
            - NSS.GHRR.NL.D01035.S2342.E0135.B0192627.WI

        Args:
            max_diff_from_t0_head (int): Threshold for offset estimation: A
                scanline timestamp matches the header timestamp t0_head if it is
                within the interval

                    [t0_head - max_diff_from_t0_head,
                    t0_head + max_diff_from_t0_head]

                around the header timestamp.
            min_frac_near_t0_head (float): Specifies the minimum fraction of
                scanline timestamps matching the header timestamp required for
                applying the correction.
            max_diff_from_ideal_t (float): Threshold for timestamp correction:
                If a scanline timestamp deviates more than max_diff_from_ideal_t
                from the ideal timestamp, it is regarded as corrupt and will be
                replaced with the ideal timestamp.

        Returns:
            Intermediate and final results (for plotting purpose)

        """
        results = {}
        dt64_msec = ">M8[ms]"

        # Check whether scanline number increases monotonically
        nums = self.scans["scan_line_number"]
        results.update({"t": self._times_as_np_datetime64.copy(), "n": nums})
        if np.any(np.diff(nums) < 0):
            LOG.error("Cannot perform timestamp correction. Scanline number does not increase monotonically.")
            results["fail_reason"] = "Scanline number jumps backwards"
            return results

        # Convert time to milliseconds since 1970-01-01
        t = self._times_as_np_datetime64.astype("i8")
        try:
            t0_head = np.array([self.get_header_timestamp().isoformat()], dtype="datetime64[ms]").astype("i8")[0]
        except ValueError as err:
            LOG.error("Cannot perform timestamp correction: %s", err)
            return

        # Compute ideal timestamps based on the scanline number. Still
        # without offset, i.e. scanline 0 has timestamp 1970-01-01 00:00
        tn = self.lineno2msec(nums)

        # Try to determine the timestamp t0 of the first scanline. Since none
        # of the actual timestamps is trustworthy, use the header timestamp
        # as a guideline. However, the header timestamp may also be corrupted,
        # so we only apply corrections if there is a minimum fraction of
        # scanlines whose timestamps match the header timestamp.
        #
        # 1) Compute offsets between actual timestamps and idealized timestamps
        offsets = t - tn

        # 2) If the offsets of a certain minimum fraction of scanlines are
        #    within a certain interval around the header timestamp, estimate
        #    t0 by calculating the median offset among these timestamps. If not,
        #    we do not have reliable information and cannot proceed.
        near_t0_head = np.where(np.fabs(offsets - t0_head) <= max_diff_from_t0_head)[0]
        results.update({"offsets": offsets, "t0_head": t0_head, "max_diff_from_t0_head": max_diff_from_t0_head})
        if near_t0_head.size / float(nums.size) >= min_frac_near_t0_head:
            t0 = np.median(offsets[near_t0_head])
        else:
            results["fail_reason"] = "Timestamp mismatch"
            raise TimestampMismatch("Timestamp mismatch. Cannot perform timestamp correction.")

        # Add estimated offset to the ideal timestamps
        tn += t0

        # Replace timestamps deviating more than a certain threshold from the
        # ideal timestamp with the ideal timestamp.
        corrupt_lines = np.where(np.fabs(t - tn) > max_diff_from_ideal_t)
        self._times_as_np_datetime64[corrupt_lines] = tn[corrupt_lines].astype(dt64_msec)
        LOG.debug("Corrected %s timestamp(s)", str(len(corrupt_lines[0])))

        return self._times_as_np_datetime64

    @property
    @abstractmethod
    def tsm_affected_intervals(self):  # pragma: no cover
        """Specify time intervals being affected by the scan motor problem.

        Returns:
            dict: Affected time intervals. A dictionary containing a list of
                (start, end) tuples for each affected platform. Both start and
                end must be datetime.datetime objects.

        """
        raise NotImplementedError

    def is_tsm_affected(self):
        """Determine whether this orbit is affected by the scan motor problem.

        Returns:
            bool: True if the orbit is affected, False otherwise.

        """
        self.get_times()
        ts, te = self.to_datetime(self._times_as_np_datetime64[[0, -1]])
        try:
            for interval in self.tsm_affected_intervals[self.spacecraft_id]:
                if ts >= interval[0] and te <= interval[1]:
                    # Found a matching interval
                    return True

            # No matching interval, orbit is not affected
            return False
        except KeyError:
            # Platform is not affected at all
            return False

    def get_midnight_scanline(self):
        """Find the scanline where the UTC date increases by one day.

        Returns:
            int: The midnight scanline if it exists and is unique.
                 None, else.

        """
        self.get_times()
        d0 = np.datetime64(datetime.date(1970, 1, 1), "D")
        days = (self._times_as_np_datetime64.astype("datetime64[D]") - d0).astype(int)
        incr = np.where(np.diff(days) == 1)[0]
        if len(incr) != 1:
            if len(incr) > 1:
                LOG.warning("Unable to determine midnight scanline: UTC date increases more than once. ")
            return None
        else:
            return incr[0]

    def get_miss_lines(self):
        """Find missing scanlines.

        I.e. scanlines which were dropped for some reason or were never recorded.

        Returns:
            Indices of missing scanlines

        """
        # Compare scanline number against the ideal case (1, 2, 3, ...) and
        # find the missing line numbers.
        ideal = set(range(1, self.scans["scan_line_number"][-1] + 1))
        missing = sorted(ideal.difference(set(self.scans["scan_line_number"])))
        return np.array(missing, dtype=int)

    def mask_tsm_pixels(self, ds):
        """Mask pixels affected by the scan motor issue."""
        idx = self.get_tsm_pixels(ds["channels"].values)
        # This is because fancy/pixel indexing doesn't seem to work with xarray's `where` or `loc`
        ds["channels"].data[idx] = np.nan

    @abstractmethod
    def get_tsm_pixels(self, channels):  # pragma: no cover
        """Determine pixels affected by the scan motor issue.

        Channel selection is POD/KLM specific.
        """
        raise NotImplementedError

    def get_attitude_coeffs(self):
        """Return the roll, pitch, yaw values."""
        if self._rpy is None:
            if "constant_yaw_attitude_error" in self.head.dtype.fields:
                rpy = np.deg2rad(
                    [
                        self.head["constant_roll_attitude_error"] / 1e3,
                        self.head["constant_pitch_attitude_error"] / 1e3,
                        self.head["constant_yaw_attitude_error"] / 1e3,
                    ]
                )
            else:
                try:
                    # This needs to be checked thoroughly first
                    # rpy_spacecraft = rpy_coeffs[self.spacecraft_name]
                    # rpy = np.array([rpy_spacecraft['roll'],
                    #                 rpy_spacecraft['pitch'],
                    #                 rpy_spacecraft['yaw']])
                    # LOG.debug("Using static attitude correction")
                    raise KeyError
                except KeyError:
                    LOG.debug("Not applying attitude correction")
                    rpy = np.zeros(3)
            LOG.info("Using rpy: %s", str(rpy))
            self._rpy = rpy
        return self._rpy

    def _compute_lonlat_from_tles(self, utcs):
        """Compute lon lat values using pyorbital."""
        tic = datetime.datetime.now()
        sgeom = self.geoloc_definition(utcs.astype(datetime.datetime), self.lonlat_sample_points)
        t0 = utcs[0].astype(datetime.datetime)
        s_times = sgeom.times(t0)
        tle1, tle2 = self.get_tle_lines()

        rpy = self.get_attitude_coeffs()
        LOG.debug(f"Computing lon/lats with attitude {rpy}")
        pixels_pos = compute_pixels((tle1, tle2), sgeom, s_times, rpy)
        pos_time = get_lonlatalt(pixels_pos, s_times)

        lons, lats = pos_time[:2]

        pixels_per_line = len(self.lonlat_sample_points)
        lons = lons.reshape(-1, pixels_per_line)
        lats = lats.reshape(-1, pixels_per_line)
        # todo: adjust time using build in lon lats
        toc = datetime.datetime.now()
        LOG.warning("Computation of geolocation: %s", str(toc - tic))

        return lons, lats


def inherit_doc(cls):
    """Make a class method inherit its docstring from the parent class.

    Copied from http://stackoverflow.com/a/8101598/5703449 .
    """
    for name, func in vars(cls).items():
        if isinstance(func, types.FunctionType) and not func.__doc__:
            for parent in cls.__bases__:
                parfunc = getattr(parent, name, None)
                if parfunc and getattr(parfunc, "__doc__", None):
                    func.__doc__ = parfunc.__doc__
                    break
    return cls



import gzip
import io
import logging
from contextlib import contextmanager, nullcontext

import numpy as np

LOG = logging.getLogger(__name__)


def gzip_inspected(open_file):
    """Try to gzip decompress the file object if applicable."""
    try:
        file_object = gzip.GzipFile(mode="rb", fileobj=open_file)
        file_object.read(1)
    except OSError:
        file_object = open_file
    finally:
        file_object.seek(0)
    return file_object


@contextmanager
def file_opener(file):
    if isinstance(file, io.IOBase) and file.seekable():
        # avoid closing the file using nullcontext
        open_file = nullcontext(file)
    elif hasattr(file, "open"):
        try:
            open_file = file.open(mode="rb")
        except TypeError:
            open_file = file.open()
    else:
        open_file = open(file, mode="rb")
    # set open_file into context in case of lazy loading in __enter__ method.
    with open_file as file_object:
        yield gzip_inspected(file_object)


def get_absolute_azimuth_angle_diff(sat_azi, sun_azi):
    """Calculates absolute azimuth difference angle. """
    rel_azi = abs(sat_azi - sun_azi)
    rel_azi = rel_azi % 360
    # Not using np.where to avoid copying array
    rel_azi[rel_azi > 180] = 360.0 - rel_azi[rel_azi > 180]
    return rel_azi


def centered_modulus(array, divisor):
    """Transform array to half open range ]-divisor/2, divisor/2]."""
    arr = array % divisor
    arr[arr > divisor / 2] -= divisor
    return arr


def calculate_sun_earth_distance_correction(jday):
    """Calculate the sun earth distance correction.

    In 2008 3-4 different equations of ESD were considered.
    This one was chosen as it at the time gave reflectances most closely
    matching the PATMOS-x data provided then by Andy Heidinger.

    Formula might need to be reconsidered if jday is updated to a float.

    """
    # Earth-Sun distance correction factor
    corr = 1.0 - 0.0334 * np.cos(2.0 * np.pi * (jday - 2) / 365.25)
    return corr


def check_user_scanlines(start_line, end_line, first_valid_lat=None,
                         last_valid_lat=None, along_track=None):
    """Check user-defined scanlines.

    Can be used by both pygac and satpy.

    Args:
        start_line: User-defined start line (afer stripping, if enabled)
        end_line: User-defined end line (afer stripping, if enabled)
        first_valid_lat: First scanline with valid latitudes
        last_valid_lat: Last scanline with valid latitudes
        along_track: Number of scanlines (only needed if stripping
            is disabled)
    """
    if first_valid_lat is not None and last_valid_lat is not None:
        num_valid_lines = last_valid_lat - first_valid_lat + 1
    else:
        if along_track is None:
            raise ValueError("Need along_track")
        num_valid_lines = along_track

    start_line = int(start_line)
    end_line = int(end_line)
    if end_line == 0:
        # If the user specifies 0 as the last scanline, process all
        # scanlines with valid coordinates
        end_line = num_valid_lines - 1
    elif end_line >= num_valid_lines:
        end_line = num_valid_lines - 1
        LOG.warning("Given end line exceeds scanline range, resetting "
                    "to {}".format(end_line))
    if start_line > num_valid_lines:
        raise ValueError("Given start line {} exceeds scanline range {}"
                         .format(start_line, num_valid_lines))
    return start_line, end_line


def strip_invalid_lat(lats):
    """Strip invalid latitudes at the end and beginning of the orbit."""
    no_wrong_lat = np.where(np.logical_not(np.isnan(lats)))
    return min(no_wrong_lat[0]), max(no_wrong_lat[0])


def slice_channel(ch, start_line, end_line, first_valid_lat=None,
                  last_valid_lat=None, midnight_scanline=None,
                  miss_lines=None, qual_flags=None):
    """Slice channel data using user-defined start/end line.

    If valid_lat_start/end are given, strip scanlines with invalid
    coordinates at the beginning and end of the orbit.

    Can be used by both pygac and satpy.

    Args:
        ch: Channel data
        start_line: User-defined start line (afer stripping, if enabled)
        end_line: User-defined end line (after stripping, if enabled)
        first_valid_lat: First scanline with valid latitudes
        last_valid_lat: Last scanline with valid latitudes.
        midnight_scanline: If given, update midnight scanline to the new
            scanline range.
        miss_lines: If given, update list of missing lines with the ones
            that have been stripped due to invalid coordinates
        qual_flags: Quality flags, needed to updated missing lines.
    """
    if first_valid_lat is not None and last_valid_lat is not None:
        # Strip invalid coordinates and update midnight scanline as well as
        # user-defined start/end lines
        ch, updated = _slice(ch,
                             start_line=first_valid_lat,
                             end_line=last_valid_lat,
                             update=[midnight_scanline])
        midnight_scanline = updated[0]

        # Reset user-defined end line, if it has been removed
        end_line = min(end_line, ch.shape[0] - 1)
        start_line = min(start_line, ch.shape[0] - 1)

        # Update missing scanlines
        if miss_lines is not None:
            miss_lines = _update_missing_scanlines(
                miss_lines=miss_lines,
                qual_flags=qual_flags,
                start_line=first_valid_lat,
                end_line=last_valid_lat)

    # Slice data using user-defined start/end lines
    ch_slc, updated = _slice(ch, start_line=start_line, end_line=end_line,
                             update=[midnight_scanline])
    midnight_scanline = updated[0]

    return ch_slc, miss_lines, midnight_scanline


def _slice(ch, start_line, end_line, update=None):
    """Slice the given channel.

    Args:
        start_line: New start line
        end_line: New end line
        update: List of scanlines to be updated to the new range
    """
    # Slice data using new start/end lines
    if len(ch.shape) == 1:
        ch_slc = ch[start_line:end_line + 1].copy()
    else:
        ch_slc = ch[start_line:end_line + 1, :].copy()

    if update:
        updated = [_update_scanline(line, start_line, end_line)
                   if line is not None else None
                   for line in update]
        return ch_slc, updated

    return ch_slc


def _update_scanline(scanline, new_start_line, new_end_line):
    """Update the given scanline to the new range.

    Set scanline to None if it lies outside the new range.
    """
    scanline -= new_start_line
    num_lines = new_end_line - new_start_line + 1
    if scanline < 0 or scanline >= num_lines:
        scanline = None
    return scanline


def _update_missing_scanlines(miss_lines, qual_flags, start_line, end_line):
    """Add scanlines excluded by slicing to the list of missing scanlines.

    Args:
        miss_lines: List of missing scanlines
        qual_flags: Quality flags
        start_line: New start line of the slice
        end_line: New end line of the slice
    """
    return np.sort(np.unique(
        qual_flags[0:start_line, 0].tolist() +
        miss_lines.tolist() +
        qual_flags[end_line + 1:, 0].tolist()
    ))


def plot_correct_times_thresh(res, filename=None):
    """Visualize results of GACReader.correct_times_thresh."""
    import matplotlib.pyplot as plt

    t = res["t"]
    tcorr = res.get("tcorr")
    n = res["n"]
    offsets = res.get("offsets")
    t0_head = res.get("t0_head")
    max_diff_from_t0_head = res.get("max_diff_from_t0_head")
    fail_reason = res.get("fail_reason", "Failed for unknown reason")

    # Setup figure
    along_track = np.arange(t.size)
    _, (ax0, ax1, ax2) = plt.subplots(nrows=3, sharex=True,
                                      figsize=(8, 10))

    # Plot original vs corrected timestamps
    ax0.plot(along_track, t, "b-", label="original")
    if tcorr is not None:
        ax0.plot(along_track, tcorr, color="red", linestyle="--",
                 label="corrected")
    else:
        ax0.set_title(fail_reason)

    ax0.set_ylabel("Time")
    ax0.set_ylim(t.min() - np.timedelta64(30, "m"),
                 t.max() + np.timedelta64(30, "m"))
    ax0.legend(loc="best")

    # Plot offset (original time - ideal time)
    if offsets is not None:
        ax1.plot(along_track, offsets)
        ax1.fill_between(
            along_track,
            t0_head - np.ones(along_track.size) * max_diff_from_t0_head,
            t0_head + np.ones(along_track.size) * max_diff_from_t0_head,
            facecolor="g", alpha=0.33)
        ax1.axhline(y=t0_head, color="g", linestyle="--",
                    label="Header timestamp")
        ax1.set_ylim(t0_head - 5 * max_diff_from_t0_head,
                     t0_head + 5 * max_diff_from_t0_head)
        ax1.set_ylabel("Offset t-tn [ms]")
        ax1.legend(loc="best")

    # Plot scanline number
    ax2.plot(along_track, n)
    ax2.set_ylabel("Scanline number")
    ax2.set_xlabel("Along Track")

    if filename:
        plt.savefig(filename, bbox_inches="tight", dpi=100)
    else:
        plt.show()


def plot_correct_scanline_numbers(res, filename=None):
    """Visualize results of GACReader.correct_scanline_numbers."""
    import matplotlib.pyplot as plt

    along_track = res["along_track"]
    n_orig = res["n_orig"]
    n_corr = res["n_corr"]
    within_range = res["within_range"]
    thresh = res["thresh"]
    diffs = res["diffs"]
    nz_diffs = res["nz_diffs"]

    # Setup figure
    _, (ax0, ax1) = plt.subplots(nrows=2)

    # Plot original vs corrected scanline numbers
    ax0.plot(along_track, n_orig, "b-", label="original")
    along_track_corr = along_track.copy()
    along_track_corr = along_track_corr[within_range]
    along_track_corr = along_track_corr[diffs <= thresh]
    ax0.plot(along_track_corr, n_corr, "r--", label="corrected")
    ax0.set_ylabel("Scanline Number")
    ax0.set_xlabel("Along Track")
    ax0.legend(loc="best")

    # Plot difference from ideal
    ax1.plot(np.arange(len(nz_diffs)), nz_diffs)
    ax1.axhline(thresh, color="r", label="thresh={0:.2f}"
                .format(thresh))
    ax1.set_xlabel("Index")
    ax1.set_ylabel("nonzero |n - n'|")
    ax1.legend()

    plt.tight_layout()

    if filename:
        plt.savefig(filename, bbox_inches="tight")
    else:
        plt.show()



"""Reader for LAC POD data."""

import logging

import numpy as np

from pygac.lac_reader import LACReader
from pygac.pod_reader import PODReader, main_pod

LOG = logging.getLogger(__name__)

scanline = np.dtype([("scan_line_number", ">i2"),
                     ("time_code", ">u2", (3, )),
                     ("quality_indicators", ">u4"),
                     ("calibration_coefficients", ">i4", (10, )),
                     ("number_of_meaningful_zenith_angles_and_earth_location_appended", ">u1"),
                     ("solar_zenith_angles", "i1", (51, )),
                     ("earth_location", [("lats", ">i2"),
                                         ("lons", ">i2")], (51,)),
                     ("telemetry", ">u4", (35, )),
                     ("sensor_data", ">u4", (3414, )),
                     ("add_on_zenith", ">u2", (10, )),
                     ("clock_drift_delta", ">u2"),  # only in newest version
                     ("spare3", "u2", (337, ))])


class LACPODReader(LACReader, PODReader):
    """The LAC POD reader.

    The `scan_points` attributes provides the position of the longitude and latitude points to
    compute relative to the full swath width.

    The offset attribute tells where in the file the scanline data starts.
    """

    def __init__(self, *args, **kwargs):
        """Init the LAC POD reader."""
        LACReader.__init__(self, *args, **kwargs)
        self.scanline_type = scanline
        self.offset = 14800
        self.scan_points = np.arange(2048)


def main(filename, start_line, end_line):
    """Generate a l1c file."""
    return main_pod(LACPODReader, filename, start_line, end_line)


if __name__ == "__main__":
    import sys
    main(sys.argv[1], sys.argv[2], sys.argv[3])



"""Test function for the calibration coeffictions handling.
"""

import sys
import unittest
from unittest import mock

import numpy as np

from pygac.calibration.noaa import Calibrator, CoeffStatus, calibrate_solar

# dummy user json file including only noaa19 data with changed channel 1 coefficients
user_json_file = b"""{
    "noaa19": {
        "channel_1": {
            "dark_count": 0,
            "gain_switch": 1000,
            "s0": 2,
            "s1": 0,
            "s2": 0
        },
        "channel_2": {
            "dark_count": 39.0,
            "gain_switch": 500.37,
            "s0": 0.122,
            "s1": 0.95,
            "s2": -0.039
        },
        "channel_3a": {
            "dark_count": 39.4,
            "gain_switch": 496.11,
            "s0": 0.1,
            "s1": 0.0,
            "s2": 0.0
        },
        "channel_3b": {
            "b0": 0.0,
            "b1": 0.0,
            "b2": 0.0,
            "centroid_wavenumber": 2670.2425,
            "space_radiance": 0.0,
            "to_eff_blackbody_intercept": 1.6863857,
            "to_eff_blackbody_slope": 0.9974112191806167
        },
        "channel_4": {
            "b0": 5.7,
            "b1": -0.11187000000000002,
            "b2": 0.00054668,
            "centroid_wavenumber": 927.92374,
            "space_radiance": -5.49,
            "to_eff_blackbody_intercept": 0.39419031,
            "to_eff_blackbody_slope": 0.9986718662850276
        },
        "channel_5": {
            "b0": 3.58,
            "b1": -0.05991000000000002,
            "b2": 0.00024985,
            "centroid_wavenumber": 831.28619,
            "space_radiance": -3.39,
            "to_eff_blackbody_intercept": 0.2636462,
            "to_eff_blackbody_slope": 0.9990463103920997
        },
        "date_of_launch": "2009-02-05T00:57:36.000000Z",
        "thermometer_1": {
            "d0": 276.6067,
            "d1": 0.051111,
            "d2": 1.405783e-06,
            "d3": 0.0,
            "d4": 0.0
        },
        "thermometer_2": {
            "d0": 276.6119,
            "d1": 0.05109,
            "d2": 1.496037e-06,
            "d3": 0.0,
            "d4": 0.0
        },
        "thermometer_3": {
            "d0": 276.6311,
            "d1": 0.051033,
            "d2": 1.49699e-06,
            "d3": 0.0,
            "d4": 0.0
        },
        "thermometer_4": {
            "d0": 276.6268,
            "d1": 0.051058,
            "d2": 1.49311e-06,
            "d3": 0.0,
            "d4": 0.0
        }
    }
}"""


class TestCalibrationCoefficientsHandling(unittest.TestCase):

    @mock.patch("pygac.calibration.noaa.open", mock.mock_open(read_data=user_json_file))
    def test_user_coefficients_file(self):
        if sys.version_info.major < 3:
            cal = Calibrator("noaa19", coeffs_file="/path/to/unknow/defaults.json")
        else:
            with self.assertWarnsRegex(RuntimeWarning,
                                       "Unknown calibration coefficients version!"):
                cal = Calibrator("noaa19", coeffs_file="/path/to/unknow/defaults.json")

        self.assertEqual(cal.dark_count[0], 0)
        self.assertEqual(cal.gain_switch[0], 1000)
        self.assertEqual(cal.s0[0], 2)
        self.assertEqual(cal.s1[0], 0)
        self.assertEqual(cal.s2[0], 0)
        # check that the version is set to None if an unknown file is used
        if sys.version_info.major > 2:
            self.assertIsNone(cal.version)

    def test_custom_coefficients(self):
        custom_coeffs = {
            "channel_1": {
                "dark_count": 0,
                "gain_switch": 1000,
                "s0": 2,
                "s1": 0,
                "s2": 0
            }
        }
        # The coefficients are choosen to preserve the counts of channel 1 if counts are less than 1000
        counts = np.arange(10)
        year = 2010
        jday = 1
        spacecraft_id = "noaa19"
        channel = 0
        cal = Calibrator(spacecraft_id, custom_coeffs=custom_coeffs)
        scaled_radiance = calibrate_solar(counts, channel, year, jday, cal)
        np.testing.assert_allclose(scaled_radiance, counts)
        # check that the version is set to None if custom coeffs are used
        if sys.version_info.major > 2:
            self.assertIsNone(cal.version)

    def test_vis_deprecation_warning(self):
        counts = np.arange(10)
        year = 2010
        jday = 1
        spacecraft_id = "noaa19"
        channel = 0
        corr = 2
        message = (
            "Using the 'corr' argument is depricated in favor of making the units"
            " of the function result clear. Please make any unit conversion outside this function."
        )
        cal = Calibrator(spacecraft_id)
        with self.assertWarnsRegex(DeprecationWarning, message):
            calibrate_solar(counts, channel, year, jday, cal, corr=corr)
        # check that the version is set in this case
        self.assertIsNotNone(cal.version)

    def test_default_coeffs(self):
        """Test identification of default coefficients."""
        _, version = Calibrator.read_coeffs(None)
        self.assertIsNotNone(version)

    def test_read_coeffs_warnings(self):
        """Test warnings issued by Calibrator.read_coeffs."""
        version_dicts = [
            # Non-nominal coefficients
            {"name": "v123",
             "status": CoeffStatus.PROVISIONAL},
            # Unknown coefficients
            {"name": None,
             "status": None}
        ]
        with mock.patch.object(Calibrator, "version_hashs") as version_hashs:
            for version_dict in version_dicts:
                version_hashs.get.return_value = version_dict
                with self.assertWarns(RuntimeWarning):
                    Calibrator.read_coeffs(None)



"""Processing utilities for GAC/LAC KLM/POD files.

Functions:
    process_file: allows to process a given input file
    get_reader_class: allows to select the appropriate reader
                      class for a given file.
"""

import datetime
import logging
import os

from pygac.configuration import get_config
from pygac.gac_klm import GACKLMReader
from pygac.gac_pod import GACPODReader
from pygac.lac_klm import LACKLMReader
from pygac.lac_pod import LACPODReader
from pygac.utils import file_opener

LOG = logging.getLogger(__name__)

_reader_classes = [GACKLMReader, LACKLMReader, GACPODReader, LACPODReader]


def get_reader_class(filename, fileobj=None):
    """Return the reader class that can read the GAC/LAC KLM/POD file.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)
    """
    found_reader = False
    for index, Reader in enumerate(_reader_classes):
        if Reader.can_read(filename, fileobj=fileobj):
            LOG.debug("%s can read the file." % Reader.__name__)
            found_reader = True
            break
    if not found_reader:
        raise ValueError('Unable to read the file "%s"' % filename)
    # Move the Reader in front of _reader_classes. Chance is high that the
    # next file is of the same kind.
    _reader_classes.insert(0, _reader_classes.pop(index))
    return Reader


def process_file(filename, start_line, end_line, fileobj=None):
    """Read, calibrate and navigate NOAA AVHRR GAC/LAC POD/KLM data.
    It creates three hdf5 files in the output location given by the pygac
    config file. The three files contain the avhrr data, quality flags,
    and sunsatangles.

    Argsuments
        filename (str): Path to GAC/LAC file
        start_line (int): First scanline to be processed (0-based)
        end_line (int): Last scanline to be processed (0-based),
                        set to 0 for the last available scanline
        fileobj: An open file object to read from. (optional)

    Note
        This function expects an initialized config file.
    """
    tic = datetime.datetime.now()
    LOG.info("Process file: %s", str(filename))

    # reader specific values
    config = get_config()
    tle_dir = config.get("tle", "tledir", raw=True)
    tle_name = config.get("tle", "tlename", raw=True)
    coeffs_file = config.get("calibration", "coeffs_file", fallback="")
    # output specific values
    output_dir = config.get("output", "output_dir", raw=True)
    output_file_prefix = config.get("output", "output_file_prefix", raw=True)
    avhrr_dir = os.environ.get("SM_AVHRR_DIR")
    qual_dir = os.environ.get("SM_AVHRR_DIR")
    sunsatangles_dir = os.environ.get("SM_SUNSATANGLES_DIR")

    # Keep the file open while searching for the reader class and later
    # creation of the instance.
    with file_opener(fileobj or filename) as open_file:
        reader_cls = get_reader_class(filename, fileobj=open_file)
        reader = reader_cls(
            tle_dir=tle_dir, tle_name=tle_name,
            calibration_file=coeffs_file
        )
        reader.read(filename, fileobj=fileobj)
        reader.save(
            start_line, end_line,
            output_file_prefix=output_file_prefix,
            output_dir=output_dir,
            avhrr_dir=avhrr_dir, qual_dir=qual_dir,
            sunsatangles_dir=sunsatangles_dir
        )
    LOG.info("Processing took: %s", str(datetime.datetime.now() - tic))



"""Test utilities."""

import numpy as np


class CalledWithArray(object):
    """Adapter for arrays in mock.assert_called_with()."""

    def __init__(self, array):
        self.array = array

    def __repr__(self):
        return repr(self.array)

    def __eq__(self, other):
        return np.all(self.array == other)



"""Test the I/O."""

import unittest
from unittest import mock

import numpy as np
import numpy.testing

import pygac.gac_io as gac_io
import pygac.utils as utils


class TestIO(unittest.TestCase):
    """Test the gac_io module."""

    longMessage = True

    def test_strip_invalid_lat(self):
        """Test stripping the invalid lats."""
        lats = np.array([np.nan, 1, np.nan, 2, np.nan])
        start, end = utils.strip_invalid_lat(lats)
        self.assertEqual(start, 1)
        self.assertEqual(end, 3)

    def test_update_scanline(self):
        """Test updating the scanlines."""
        test_data = [{"new_start_line": 100, "new_end_line": 200,
                      "scanline": 110, "scanline_exp": 10},
                     {"new_start_line": 100, "new_end_line": 200,
                      "scanline": 90, "scanline_exp": None},
                     {"new_start_line": 100, "new_end_line": 200,
                      "scanline": 210, "scanline_exp": None}]
        for t in test_data:
            scanline_exp = t.pop("scanline_exp")
            scanline = utils._update_scanline(**t)
            self.assertEqual(scanline, scanline_exp)

    def test_update_missing_scanlines(self):
        """Test updating the missing scanlines."""
        qual_flags = np.array([[1, 2, 4, 5, 6, 8, 9, 11, 12]]).transpose()
        miss_lines = np.array([3, 7, 10])
        test_data = [{"start_line": 0, "end_line": 8,
                      "miss_lines_exp": [3, 7, 10]},
                     {"start_line": 3, "end_line": 6,
                      "miss_lines_exp": [1, 2, 3, 4, 7, 10, 11, 12]}]
        for t in test_data:
            miss_lines_exp = t.pop("miss_lines_exp")
            miss_lines = utils._update_missing_scanlines(
                miss_lines=miss_lines, qual_flags=qual_flags, **t)
            numpy.testing.assert_array_equal(miss_lines, miss_lines_exp)

        # If intersection of miss_lines and qual_flags is not empty
        # (here: extra "1" in miss_lines), make sure that items are
        # not added twice.
        miss_lines = utils._update_missing_scanlines(
            miss_lines=np.array([1, 3, 7, 10]),
            qual_flags=qual_flags,
            start_line=3, end_line=6)
        numpy.testing.assert_array_equal(miss_lines,
                                         [1, 2, 3, 4, 7, 10, 11, 12])

    def test_slice(self):
        """Test slices."""
        ch = np.array([[1, 2, 3, 4, 5]]).transpose()
        sliced_exp = np.array([[2, 3, 4]]).transpose()

        # Without update
        sliced = utils._slice(ch, start_line=1, end_line=3)
        numpy.testing.assert_array_equal(sliced, sliced_exp)

        # With update
        sliced, updated = utils._slice(ch, start_line=1, end_line=3,
                                       update=[0, 2, 4, None])
        numpy.testing.assert_array_equal(sliced, sliced_exp)
        numpy.testing.assert_array_equal(updated, [None, 1, None, None])

        # Make sure slice is a copy
        ch += 1
        numpy.testing.assert_array_equal(sliced, sliced_exp)

    def test_slice_channel(self):
        """Test selection of user defined scanlines.

        Scanline Nr:  1  2  3  4  5  6  7  8  9 10 11 12 13 14
        Missing:      X        X                    X  X
        Inv. lat/lon:    X  X                             X  X

        Before stripping of invalid lats
        --------------------------------
        idx:             0  1     2  3  4  5  6  7        8  9
        data:            1  2     3  4  5  6  7  8        9 10

        After stripping of invalid lats
        -------------------------------
        idx:                      0  1  2  3  4  5
        data:                     3  4  5  6  7  8


        => All missing lines: [1, 2, 3, 4, 11, 12, 13, 14]
        """
        # Define test data
        ch = np.array([[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]]).transpose()
        miss_lines = np.array([1, 4, 11, 12])  # never recorded
        qualflags = np.array(
            [[2, 3, 5, 6, 7, 8, 9, 10, 13, 14]]).transpose()  # scanline number
        midn_line = 3
        first_valid_lat = 2
        last_valid_lat = 7
        start_line = 1
        end_line = 4

        # Without lat stripping
        sliced_ref = np.array([[2, 3, 4, 5]]).transpose()
        sliced, miss_lines_new, midn_line_new = utils.slice_channel(
            ch, start_line=start_line, end_line=end_line,
            miss_lines=miss_lines, midnight_scanline=midn_line,
            qual_flags=qualflags)
        numpy.testing.assert_array_equal(sliced, sliced_ref)
        numpy.testing.assert_array_equal(miss_lines_new, miss_lines)
        self.assertEqual(midn_line_new, 2)

        # With lat stripping
        sliced_ref = np.array([[4, 5, 6, 7]]).transpose()
        miss_lines_ref = np.array([1, 2, 3, 4, 11, 12, 13, 14])
        sliced, miss_lines_new, midn_line_new = utils.slice_channel(
            ch, start_line=start_line, end_line=end_line,
            first_valid_lat=first_valid_lat, last_valid_lat=last_valid_lat,
            miss_lines=miss_lines, midnight_scanline=midn_line,
            qual_flags=qualflags)
        numpy.testing.assert_array_equal(sliced, sliced_ref)
        numpy.testing.assert_array_equal(miss_lines_new, miss_lines_ref)
        self.assertEqual(midn_line_new, 0)

    def test_check_user_scanlines(self):
        """Check the scanlines."""
        # All scanlines
        start, end = utils.check_user_scanlines(0, 0, 100, 200)
        self.assertEqual(start, 0)
        self.assertEqual(end, 100)

        start, end = utils.check_user_scanlines(0, 0, None, None, 100)
        self.assertEqual(start, 0)
        self.assertEqual(end, 99)

        # Valid scanlines
        start, end = utils.check_user_scanlines(10, 20, 100, 200)
        self.assertEqual(start, 10)
        self.assertEqual(end, 20)

        start, end = utils.check_user_scanlines(10, 20, None, None, 100)
        self.assertEqual(start, 10)
        self.assertEqual(end, 20)

        # Invalid scanlines
        start, end = utils.check_user_scanlines(10, 110, 100, 200)
        self.assertEqual(start, 10)
        self.assertEqual(end, 100)

        start, end = utils.check_user_scanlines(10, 110, None, None, 100)
        self.assertEqual(start, 10)
        self.assertEqual(end, 99)

        self.assertRaises(ValueError, gac_io.check_user_scanlines,
                          110, 120, 100, 200)
        self.assertRaises(ValueError, gac_io.check_user_scanlines,
                          110, 120, None, None, 100)

    @mock.patch("pygac.gac_io.strip_invalid_lat")
    @mock.patch("pygac.gac_io.avhrrGAC_io")
    @mock.patch("pygac.gac_io.slice_channel")
    @mock.patch("pygac.gac_io.check_user_scanlines")
    def test_save_gac(self, check_user_scanlines, slice_channel, avhrr_gac_io,
                      strip_invalid_lat):
        """Test saving."""
        # Test scanline selection
        mm = mock.MagicMock()
        kwargs = dict(
            satellite_name=mm,
            xutcs=mm,
            lats=mm,
            lons=mm,
            ref1=mm,
            ref2=mm,
            ref3=mm,
            bt3=mm,
            bt4=mm,
            bt5=mm,
            sun_zen=mm,
            sat_zen=mm,
            sun_azi=mm,
            sat_azi=mm,
            rel_azi=mm,
            qual_flags=mm,
            gac_file=mm,
            meta_data=mm,
            output_file_prefix=mm,
            avhrr_dir=mm,
            qual_dir=mm,
            sunsatangles_dir=mm
        )
        slice_channel.return_value = mm, "miss", "midnight"
        strip_invalid_lat.return_value = 0, 0
        check_user_scanlines.return_value = "start", "end"

        gac_io.save_gac(start_line=0, end_line=0, **kwargs)
        slice_channel.assert_called_with(mock.ANY,
                                         start_line="start", end_line="end",
                                         first_valid_lat=mock.ANY,
                                         last_valid_lat=mock.ANY
                                         )
        expected_args = [
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY,
            "midnight",
            "miss",
            mock.ANY,
            mock.ANY,
            mock.ANY,
            mock.ANY
        ]
        avhrr_gac_io.assert_called_with(*expected_args)



"""Reader for GAC POD data."""

import logging

import numpy as np

from pygac.gac_reader import GACReader
from pygac.pod_reader import PODReader, main_pod

LOG = logging.getLogger(__name__)

scanline = np.dtype([("scan_line_number", ">i2"),
                     ("time_code", ">u2", (3, )),
                     ("quality_indicators", ">u4"),
                     ("calibration_coefficients", ">i4", (10, )),
                     ("number_of_meaningful_zenith_angles_and_earth_location_appended",
                      ">u1"),
                     ("solar_zenith_angles", "i1", (51, )),
                     ("earth_location", [("lats", ">i2"),
                                         ("lons", ">i2")], (51,)),
                     ("telemetry", ">u4", (35, )),
                     ("sensor_data", ">u4", (682, )),
                     ("add_on_zenith", ">u2", (10, )),
                     ("clock_drift_delta", ">u2"),
                     ("spare3", "u2", (11, ))])


class GACPODReader(GACReader, PODReader):
    """The GAC POD reader class.

    The `scan_points` attributes provides the position of the longitude and latitude points to
    compute relative to the full swath width.

    The offset attribute tells where in the file the scanline data starts.
    """

    def __init__(self, *args, **kwargs):
        """Init the GAC POD reader."""
        GACReader.__init__(self, *args, **kwargs)
        self.scanline_type = scanline
        # GAC POD files were originally written to tapes using 6440 byte physical records
        # with two 3220 logical records per physical record. As the header is in the first
        # physical record the scanline data will start at offset 6440 in the filestream.
        # This means the second logical record (at offset 3220) is not used and contains
        # junk / padding data. The data often appears to be a valid scanline, but it is a
        # duplicate of a data appearing later in the file and should not be used.
        self.offset = 6440
        self.scan_points = np.arange(3.5, 2048, 5)


def main(filename, start_line, end_line):
    """Generate a l1c file."""
    return main_pod(GACPODReader, filename, start_line, end_line)


if __name__ == "__main__":
    import sys
    main(sys.argv[1], sys.argv[2], sys.argv[3])



"""NOAA-14, 15 and 16 suffer from temporary scan motor issue i.e. parts of
an orbit contain corrupt data. This module identifies affected pixels and flags
them with fill values."""

import datetime

import bottleneck as bn
import numpy as np

TSM_AFFECTED_INTERVALS_POD = {
    3: [(datetime.datetime(2001, 10, 19, 4, 50), datetime.datetime(2001, 10, 19, 13, 38)),
        (datetime.datetime(2001, 10, 19, 16, 58),
         datetime.datetime(2001, 10, 19, 18, 7)),
        (datetime.datetime(2001, 10, 19, 23, 55),
         datetime.datetime(2001, 10, 27, 3, 19)),
        (datetime.datetime(2001, 10, 27, 4, 53),
         datetime.datetime(2001, 10, 31, 21, 7)),
        (datetime.datetime(2001, 10, 31, 20, 58),
         datetime.datetime(2001, 11, 13, 12, 46)),
        (datetime.datetime(2001, 11, 13, 13, 25),
         datetime.datetime(2001, 11, 21, 23, 40)),
        (datetime.datetime(2001, 11, 22, 1, 14),
         datetime.datetime(2001, 11, 25, 6, 1)),
        (datetime.datetime(2001, 11, 25, 11, 3),
         datetime.datetime(2001, 11, 25, 13, 8)),
        (datetime.datetime(2001, 11, 26, 10, 51),
         datetime.datetime(2001, 11, 26, 12, 55)),
        (datetime.datetime(2001, 11, 26, 15, 49),
         datetime.datetime(2001, 11, 26, 19, 26)),
        (datetime.datetime(2001, 11, 26, 20, 36),
         datetime.datetime(2001, 11, 26, 22, 31)),
        (datetime.datetime(2001, 11, 27, 0, 5),
         datetime.datetime(2001, 12, 2, 23, 7)),
        (datetime.datetime(2001, 12, 3, 0, 41),
         datetime.datetime(2001, 12, 15, 0, 2)),
        (datetime.datetime(2001, 12, 15, 1, 30),
         datetime.datetime(2001, 12, 15, 3, 26)),
        (datetime.datetime(2001, 12, 15, 5, 2),
         datetime.datetime(2001, 12, 15, 10, 36)),
        (datetime.datetime(2001, 12, 15, 12, 2),
         datetime.datetime(2001, 12, 15, 13, 53)),
        (datetime.datetime(2001, 12, 15, 15, 21),
         datetime.datetime(2001, 12, 15, 18, 56)),
        (datetime.datetime(2001, 12, 15, 23, 37),
         datetime.datetime(2001, 12, 16, 1, 33)),
        (datetime.datetime(2002, 1, 1, 10, 14),
         datetime.datetime(2002, 1, 10, 1, 29)),
        (datetime.datetime(2002, 7, 26, 14, 40),
         datetime.datetime(2002, 7, 27, 13, 6)),
        (datetime.datetime(2002, 7, 29, 5, 26),
         datetime.datetime(2002, 8, 8, 0, 2)),
        (datetime.datetime(2002, 8, 8, 1, 32),
         datetime.datetime(2002, 8, 12, 22, 51)),
        (datetime.datetime(2002, 8, 13, 0, 25),
         datetime.datetime(2002, 9, 30, 4, 8))
        ]  # NOAA-14
}
TSM_AFFECTED_INTERVALS_KLM = {
    4: [(datetime.datetime(2001, 1, 8, 23, 55), datetime.datetime(2001, 1, 9, 23, 47)),
        (datetime.datetime(2001, 1, 10, 1, 14),
         datetime.datetime(2001, 1, 11, 1, 6)),
        (datetime.datetime(2001, 1, 23, 18, 32),
         datetime.datetime(2001, 1, 26, 0, 25)),
        (datetime.datetime(2001, 1, 30, 10, 59),
         datetime.datetime(2001, 1, 30, 17, 58)),
        (datetime.datetime(2001, 2, 2, 2, 42),
         datetime.datetime(2001, 2, 8, 11, 10)),
        (datetime.datetime(2001, 2, 11, 6, 3),
         datetime.datetime(2001, 2, 16, 21, 32)),
        (datetime.datetime(2001, 4, 24, 5, 25),
         datetime.datetime(2001, 4, 24, 9, 18)),
        (datetime.datetime(2001, 4, 24, 10, 53),
         datetime.datetime(2001, 4, 25, 8, 55)),
        (datetime.datetime(2001, 4, 26, 2, 59),
         datetime.datetime(2001, 4, 26, 8, 25)),
        (datetime.datetime(2001, 4, 26, 18, 22),
         datetime.datetime(2001, 4, 26, 20, 18)),
        (datetime.datetime(2001, 4, 26, 23, 5),
         datetime.datetime(2001, 4, 27, 8, 2)),
        (datetime.datetime(2001, 4, 28, 22, 23),
         datetime.datetime(2001, 4, 29, 0, 14)),
        (datetime.datetime(2001, 4, 29, 15, 33),
         datetime.datetime(2001, 4, 29, 17, 28)),
        (datetime.datetime(2001, 4, 29, 20, 33),
         datetime.datetime(2001, 4, 30, 8, 35)),
        (datetime.datetime(2001, 4, 30, 13, 35),
         datetime.datetime(2001, 4, 30, 18, 46)),
        (datetime.datetime(2001, 4, 30, 20, 11),
         datetime.datetime(2001, 5, 1, 8, 11)),
        (datetime.datetime(2001, 5, 2, 2, 23),
         datetime.datetime(2001, 5, 2, 5, 58)),
        (datetime.datetime(2001, 5, 2, 9, 22),
         datetime.datetime(2001, 5, 2, 22, 45)),
        (datetime.datetime(2001, 5, 3, 0, 8),
         datetime.datetime(2001, 5, 3, 9, 14)),
        (datetime.datetime(2001, 5, 8, 10, 34),
         datetime.datetime(2001, 5, 8, 14, 8)),
        (datetime.datetime(2001, 5, 8, 15, 30),
         datetime.datetime(2001, 5, 9, 15, 34)),
        (datetime.datetime(2001, 5, 10, 0, 50),
         datetime.datetime(2001, 5, 10, 4, 35)),
        (datetime.datetime(2001, 5, 10, 13, 7),
         datetime.datetime(2001, 5, 10, 21, 48)),
        (datetime.datetime(2001, 5, 10, 22, 47),
         datetime.datetime(2001, 5, 11, 0, 42)),
        (datetime.datetime(2001, 5, 11, 3, 57),
         datetime.datetime(2001, 5, 11, 7, 43)),
        (datetime.datetime(2001, 5, 11, 11, 5),
         datetime.datetime(2001, 5, 11, 17, 55)),
        (datetime.datetime(2001, 5, 11, 19, 21),
         datetime.datetime(2001, 5, 12, 7, 19)),
        (datetime.datetime(2001, 5, 12, 8, 56),
         datetime.datetime(2001, 5, 14, 6, 26)),
        (datetime.datetime(2001, 5, 14, 21, 32),
         datetime.datetime(2001, 5, 14, 23, 11)),
        (datetime.datetime(2001, 5, 15, 4, 6),
         datetime.datetime(2001, 5, 15, 7, 52)),
        (datetime.datetime(2001, 5, 15, 17, 59),
         datetime.datetime(2001, 5, 15, 19, 44)),
        (datetime.datetime(2001, 5, 15, 21, 19),
         datetime.datetime(2001, 5, 16, 7, 29)),
        (datetime.datetime(2001, 5, 16, 17, 26),
         datetime.datetime(2001, 5, 16, 19, 22)),
        (datetime.datetime(2001, 5, 16, 23, 50),
         datetime.datetime(2001, 5, 17, 8, 55)),
        (datetime.datetime(2001, 5, 17, 15, 26),
         datetime.datetime(2001, 5, 17, 17, 18)),
        (datetime.datetime(2001, 5, 17, 20, 25),
         datetime.datetime(2001, 5, 18, 6, 36)),
        (datetime.datetime(2001, 6, 5, 16, 34),
         datetime.datetime(2001, 6, 5, 21, 46)),
        (datetime.datetime(2001, 6, 6, 0, 35),
         datetime.datetime(2001, 6, 6, 7, 51)),
        (datetime.datetime(2001, 8, 4, 5, 2),
         datetime.datetime(2001, 8, 4, 8, 54)),
        (datetime.datetime(2001, 8, 6, 21, 17),
         datetime.datetime(2001, 8, 7, 7, 37)),
        (datetime.datetime(2001, 8, 7, 19, 16),
         datetime.datetime(2001, 8, 8, 12, 31)),
        (datetime.datetime(2001, 8, 8, 15, 33),
         datetime.datetime(2001, 8, 8, 22, 13)),
        (datetime.datetime(2001, 8, 23, 0, 58),
         datetime.datetime(2001, 8, 23, 15, 18)),
        (datetime.datetime(2001, 8, 23, 21, 30),
         datetime.datetime(2001, 8, 24, 7, 50)),
        (datetime.datetime(2001, 8, 24, 9, 24),
         datetime.datetime(2001, 8, 25, 0, 25)),
        (datetime.datetime(2001, 8, 25, 3, 50),
         datetime.datetime(2001, 8, 25, 11, 5)),
        (datetime.datetime(2001, 8, 25, 12, 28),
         datetime.datetime(2001, 8, 25, 17, 39)),
        (datetime.datetime(2001, 8, 25, 20, 54),
         datetime.datetime(2001, 8, 26, 8, 52)),
        (datetime.datetime(2001, 8, 26, 18, 41),
         datetime.datetime(2001, 8, 27, 4, 51)),
        (datetime.datetime(2001, 8, 27, 23, 2),
         datetime.datetime(2001, 8, 28, 0, 58)),
        (datetime.datetime(2001, 8, 28, 4, 14),
         datetime.datetime(2001, 8, 28, 7, 59)),
        (datetime.datetime(2001, 8, 28, 14, 49),
         datetime.datetime(2001, 8, 28, 16, 31)),
        (datetime.datetime(2001, 8, 28, 17, 58),
         datetime.datetime(2001, 8, 29, 17, 48)),
        (datetime.datetime(2001, 8, 29, 22, 19),
         datetime.datetime(2001, 8, 30, 0, 11)),
        (datetime.datetime(2001, 8, 30, 5, 7),
         datetime.datetime(2001, 8, 31, 13, 46)),
        (datetime.datetime(2001, 8, 31, 18, 37),
         datetime.datetime(2001, 8, 31, 22, 11)),
        (datetime.datetime(2001, 8, 31, 23, 10),
         datetime.datetime(2001, 9, 1, 8, 8)),
        (datetime.datetime(2001, 9, 1, 11, 29),
         datetime.datetime(2001, 9, 1, 13, 23)),
        (datetime.datetime(2001, 9, 1, 14, 58),
         datetime.datetime(2001, 9, 1, 16, 39)),
        (datetime.datetime(2001, 9, 1, 18, 13),
         datetime.datetime(2001, 9, 1, 23, 3)),
        (datetime.datetime(2001, 9, 2, 0, 28),
         datetime.datetime(2001, 9, 2, 7, 44)),
        (datetime.datetime(2001, 9, 2, 16, 5),
         datetime.datetime(2001, 9, 2, 18, 8)),
        (datetime.datetime(2001, 9, 2, 22, 26),
         datetime.datetime(2001, 9, 3, 9, 10)),
        (datetime.datetime(2001, 9, 3, 12, 22),
         datetime.datetime(2001, 9, 4, 5, 8)),
        (datetime.datetime(2001, 9, 7, 2, 4),
         datetime.datetime(2001, 9, 7, 9, 19)),
        (datetime.datetime(2001, 9, 7, 10, 54),
         datetime.datetime(2001, 9, 7, 19, 23)),
        (datetime.datetime(2001, 9, 7, 23, 51),
         datetime.datetime(2001, 9, 8, 20, 40)),
        (datetime.datetime(2001, 9, 8, 23, 29),
         datetime.datetime(2001, 9, 10, 22, 58)),
        (datetime.datetime(2001, 9, 11, 0, 25),
         datetime.datetime(2001, 9, 11, 4, 16)),
        (datetime.datetime(2001, 9, 11, 5, 35),
         datetime.datetime(2001, 9, 11, 12, 55)),
        (datetime.datetime(2001, 10, 25, 17, 39),
         datetime.datetime(2001, 10, 26, 0, 8)),
        (datetime.datetime(2002, 3, 17, 14, 29),
         datetime.datetime(2002, 3, 18, 2, 4)),
        (datetime.datetime(2002, 3, 18, 5, 11),
         datetime.datetime(2002, 3, 18, 7, 15)),
        (datetime.datetime(2002, 3, 18, 18, 53),
         datetime.datetime(2002, 3, 18, 20, 48))
        ],   # NOAA-15
    2: [(datetime.datetime(2004, 1, 14, 14, 13), datetime.datetime(2004, 1, 14, 22, 46)),
        (datetime.datetime(2004, 1, 15, 15, 45),
         datetime.datetime(2004, 1, 15, 19, 3)),
        (datetime.datetime(2004, 1, 15, 22, 28),
         datetime.datetime(2004, 1, 16, 2, 6)),
        (datetime.datetime(2004, 1, 16, 5, 29),
         datetime.datetime(2004, 1, 16, 22, 23)),
        (datetime.datetime(2004, 1, 16, 23, 58),
         datetime.datetime(2004, 1, 18, 0, 1)),
        (datetime.datetime(2004, 1, 18, 1, 23),
         datetime.datetime(2004, 1, 18, 8, 45)),
        (datetime.datetime(2004, 1, 18, 10, 6),
         datetime.datetime(2004, 1, 18, 23, 42)),
        (datetime.datetime(2004, 1, 19, 1, 11),
         datetime.datetime(2004, 1, 19, 15, 12)),
        (datetime.datetime(2004, 1, 19, 19, 44),
         datetime.datetime(2004, 1, 19, 23, 30)),
        (datetime.datetime(2004, 1, 20, 0, 59),
         datetime.datetime(2004, 1, 20, 4, 52)),
        (datetime.datetime(2004, 1, 20, 9, 56),
         datetime.datetime(2004, 1, 20, 15, 1)),
        (datetime.datetime(2004, 1, 20, 16, 37),
         datetime.datetime(2004, 1, 21, 4, 40)),
        (datetime.datetime(2004, 1, 21, 11, 13),
         datetime.datetime(2004, 1, 21, 16, 39)),
        (datetime.datetime(2004, 1, 22, 0, 35),
         datetime.datetime(2004, 1, 22, 6, 18)),
        (datetime.datetime(2004, 1, 22, 12, 42),
         datetime.datetime(2004, 1, 23, 2, 27)),
        (datetime.datetime(2004, 1, 23, 10, 50),
         datetime.datetime(2004, 1, 23, 22, 45)),
        (datetime.datetime(2004, 1, 24, 0, 12),
         datetime.datetime(2004, 1, 24, 5, 54)),
        (datetime.datetime(2004, 3, 16, 22, 38),
         datetime.datetime(2004, 3, 17, 22, 41)),
        (datetime.datetime(2004, 3, 18, 0, 1),
         datetime.datetime(2004, 3, 18, 5, 44)),
        (datetime.datetime(2004, 3, 20, 18, 13),
         datetime.datetime(2004, 3, 20, 23, 42)),
        (datetime.datetime(2004, 3, 21, 2, 59),
         datetime.datetime(2004, 3, 21, 5, 3)),
        (datetime.datetime(2004, 3, 21, 10, 9),
         datetime.datetime(2004, 3, 21, 23, 29)),
        (datetime.datetime(2004, 3, 22, 0, 58),
         datetime.datetime(2004, 3, 22, 6, 41)),
        (datetime.datetime(2004, 3, 22, 9, 55),
         datetime.datetime(2004, 3, 22, 15, 0)),
        (datetime.datetime(2004, 3, 22, 16, 26),
         datetime.datetime(2004, 3, 22, 23, 18)),
        (datetime.datetime(2004, 3, 23, 0, 46),
         datetime.datetime(2004, 3, 23, 4, 39)),
        (datetime.datetime(2004, 3, 23, 6, 14),
         datetime.datetime(2004, 3, 23, 11, 27)),
        (datetime.datetime(2004, 3, 23, 12, 53),
         datetime.datetime(2004, 3, 23, 16, 38)),
        (datetime.datetime(2004, 3, 23, 19, 20),
         datetime.datetime(2004, 3, 24, 6, 17)),
        (datetime.datetime(2004, 3, 24, 7, 43),
         datetime.datetime(2004, 3, 25, 6, 5)),
        (datetime.datetime(2004, 3, 25, 20, 46),
         datetime.datetime(2004, 3, 25, 22, 50)),
        (datetime.datetime(2004, 3, 26, 0, 11),
         datetime.datetime(2004, 3, 26, 5, 54)),
        (datetime.datetime(2004, 3, 26, 10, 39),
         datetime.datetime(2004, 3, 26, 14, 15)),
        (datetime.datetime(2004, 3, 28, 8, 40),
         datetime.datetime(2004, 4, 1, 23, 5)),
        (datetime.datetime(2004, 4, 2, 0, 32),
         datetime.datetime(2004, 4, 2, 22, 53)),
        (datetime.datetime(2004, 4, 3, 0, 21),
         datetime.datetime(2004, 4, 4, 22, 31)),
        (datetime.datetime(2004, 4, 5, 0, 6),
         datetime.datetime(2004, 4, 5, 22, 19)),
        (datetime.datetime(2004, 4, 5, 23, 55),
         datetime.datetime(2004, 4, 6, 5, 23)),
        (datetime.datetime(2004, 4, 6, 18, 21),
         datetime.datetime(2004, 4, 6, 23, 49)),
        (datetime.datetime(2004, 4, 7, 13, 24),
         datetime.datetime(2004, 4, 7, 15, 20)),
        (datetime.datetime(2004, 4, 7, 16, 45),
         datetime.datetime(2004, 4, 7, 23, 38)),
        (datetime.datetime(2004, 4, 8, 1, 6),
         datetime.datetime(2004, 4, 9, 10, 7)),
        (datetime.datetime(2004, 4, 21, 2, 1),
         datetime.datetime(2004, 4, 21, 17, 41)),
        (datetime.datetime(2004, 4, 21, 18, 51),
         datetime.datetime(2004, 4, 21, 22, 38)),
        (datetime.datetime(2004, 4, 22, 0, 13),
         datetime.datetime(2004, 4, 22, 5, 43)),
        (datetime.datetime(2004, 4, 24, 21, 49),
         datetime.datetime(2004, 4, 24, 23, 47)),
        (datetime.datetime(2004, 4, 25, 1, 14),
         datetime.datetime(2004, 4, 25, 20, 3)),
        (datetime.datetime(2004, 4, 26, 2, 51),
         datetime.datetime(2004, 4, 26, 11, 43)),
        (datetime.datetime(2004, 4, 26, 21, 27),
         datetime.datetime(2004, 4, 26, 23, 22)),
        (datetime.datetime(2004, 4, 27, 0, 50),
         datetime.datetime(2004, 4, 27, 4, 43)),
        (datetime.datetime(2004, 4, 27, 8, 0),
         datetime.datetime(2004, 4, 27, 10, 3)),
        (datetime.datetime(2004, 4, 27, 12, 57),
         datetime.datetime(2004, 4, 27, 21, 28)),
        (datetime.datetime(2004, 4, 28, 0, 38),
         datetime.datetime(2004, 4, 28, 4, 32)),
        (datetime.datetime(2004, 4, 28, 11, 5),
         datetime.datetime(2004, 4, 28, 16, 22)),
        (datetime.datetime(2004, 4, 29, 0, 26),
         datetime.datetime(2004, 4, 29, 14, 30)),
        (datetime.datetime(2004, 5, 2, 8, 44),
         datetime.datetime(2004, 5, 2, 23, 55)),
        (datetime.datetime(2004, 5, 3, 1, 23),
         datetime.datetime(2004, 5, 4, 3, 16)),
        (datetime.datetime(2004, 5, 4, 4, 50),
         datetime.datetime(2004, 5, 4, 8, 35)),
        (datetime.datetime(2004, 5, 4, 10, 9),
         datetime.datetime(2004, 5, 4, 23, 32)),
        (datetime.datetime(2004, 5, 5, 1, 0), datetime.datetime(2004, 5, 5, 3, 4)),
        (datetime.datetime(2004, 5, 5, 6, 27),
         datetime.datetime(2004, 5, 5, 10, 13)),
        (datetime.datetime(2004, 5, 5, 11, 27),
         datetime.datetime(2004, 5, 5, 13, 22)),
        (datetime.datetime(2004, 5, 8, 9, 23),
         datetime.datetime(2004, 5, 8, 16, 9)),
        (datetime.datetime(2004, 5, 9, 7, 22),
         datetime.datetime(2004, 5, 9, 10, 55)),
        (datetime.datetime(2004, 5, 10, 22, 8),
         datetime.datetime(2004, 5, 11, 3, 38)),
        (datetime.datetime(2004, 5, 12, 21, 44),
         datetime.datetime(2004, 5, 13, 3, 13)),
        (datetime.datetime(2004, 5, 13, 6, 37),
         datetime.datetime(2004, 5, 13, 10, 21)),
        (datetime.datetime(2004, 5, 13, 11, 35),
         datetime.datetime(2004, 5, 13, 15, 12)),
        (datetime.datetime(2004, 5, 13, 18, 2),
         datetime.datetime(2004, 5, 14, 1, 13)),
        (datetime.datetime(2004, 5, 14, 6, 25),
         datetime.datetime(2004, 5, 14, 11, 39)),
        (datetime.datetime(2004, 5, 14, 13, 5),
         datetime.datetime(2004, 5, 14, 16, 41)),
        (datetime.datetime(2004, 5, 15, 6, 14),
         datetime.datetime(2004, 5, 15, 16, 30)),
        (datetime.datetime(2004, 5, 15, 19, 22),
         datetime.datetime(2004, 5, 15, 21, 24)),
        (datetime.datetime(2004, 5, 19, 22, 9),
         datetime.datetime(2004, 5, 20, 22, 9)),
        (datetime.datetime(2004, 5, 21, 8, 28),
         datetime.datetime(2004, 5, 21, 10, 19))
        ]  # NOAA-16
}


def _rolling_window(array, size):
    """Create an array view of rolling windows.

    Args:
        array (np.ndarray): 2D array
        size (tuple): window size

    Returns:
        windows (np.ndarray): 4D array of rolling windows the first dimensions
            are the window location on the 2D array, the last two dimensions
            are the location inside the window.
    """
    shape = ((array.shape[0] - size[0] + 1, array.shape[1] - size[1] + 1,)
             + size)
    strides = array.strides + array.strides
    return np.lib.stride_tricks.as_strided(array, shape=shape, strides=strides)


def std_filter(data, box_size):
    """Filter a 2D array using a standard deviation kernel.

    Args:
        data (np.ndarray): 2D array to be filtered
        box_size (int): Specifies the boxsize. Must be odd.

    Returns:
        np.ndarray: The filtered array
    """
    size = (box_size, box_size)
    border = box_size // 2
    # need to surround the data with NaNs to calculate values at the boundary
    padded_data = np.pad(
        data, (border, border),
        mode="constant",
        constant_values=np.nan
    )
    windows = _rolling_window(padded_data, size)
    n3, n2, n1, n0 = windows.shape
    # flatten the windows for bottleneck function call
    windows = windows.reshape((n3, n2, n1*n0))
    return bn.nanstd(windows, axis=2)


def get_tsm_idx(ch1, ch2, ch4, ch5):
    """Determine indices of TSM affected pixels."""

    # absolute difference because ch1 is very similar to ch2
    abs_d12 = abs(ch1 - ch2)
    # relative difference because ch4 and ch5 differ
    rel_d45 = 100.0*(ch4 - ch5) / ch5

    # standard deviation of abs_d12 and rel_d45
    box_size = 3
    std_d12 = std_filter(abs_d12, box_size)
    std_d45 = std_filter(rel_d45, box_size)
    # replace NaNs to avoid "RuntimeWarning: invalid value encountered in greater"
    std_d12 = np.nan_to_num(std_d12, copy=False)
    std_d45 = np.nan_to_num(std_d45, copy=False)

    # using ch1, ch2, ch4, ch5 in combination
    # all channels seems to be affected throughout the whole orbit,
    # independent of VIS and NIR or day and night
    idx = np.where((std_d12 > 2.0) & (std_d45 > 2.0))

    return idx
